# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.requirement.panel.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Requirement Panels."""

# Standard Library Imports
from datetime import date
from typing import Dict, List, Tuple

# Third Party Imports
import treelib
from pubsub import pub

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gdk, Gtk, _
from ramstk.views.gtk3.widgets import (
    RAMSTKButton,
    RAMSTKCheckButton,
    RAMSTKComboBox,
    RAMSTKDateSelect,
    RAMSTKEntry,
    RAMSTKFixedPanel,
    RAMSTKTextView,
    RAMSTKTreePanel,
)


class RequirementTreePanel(RAMSTKTreePanel):
    """Panel to display hierarchy of requirements."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _select_msg = "succeed_retrieve_requirements"
    _tag = "requirement"
    _title = _("Requirement Tree")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Requirement panel."""
        super().__init__()

        # Initialize private dictionary class attributes.
        self.tvwTreeView.dic_row_loader = {
            "requirement": self.__do_load_requirement,
        }

        # Initialize private list class attributes.

        # Initialize private scalar class attributes.

        # Initialize public dictionary class attributes.
        self.dic_attribute_widget_map = {
            "revision_id": [
                0,
                Gtk.CellRendererText(),
                "edited",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Revision ID"),
                "gint",
            ],
            "requirement_id": [
                1,
                Gtk.CellRendererText(),
                "edited",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Requirement ID"),
                "gint",
            ],
            "derived": [
                2,
                Gtk.CellRendererToggle(),
                "toggled",
                super().on_cell_toggled,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Derived?"),
                "gint",
            ],
            "description": [
                3,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                "mvw_editing_requirement",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Description"),
                "gchararray",
            ],
            "figure_number": [
                4,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                "mvw_editing_requirement",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Figure Number"),
                "gchararray",
            ],
            "owner": [
                5,
                Gtk.CellRendererCombo(),
                "changed",
                super().on_cell_change,
                "mvw_editing_requirement",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Owner"),
                "gchararray",
            ],
            "page_number": [
                6,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                "mvw_editing_requirement",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Page Number"),
                "gchararray",
            ],
            "parent_id": [
                7,
                Gtk.CellRendererText(),
                "edited",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Parent ID"),
                "gint",
            ],
            "priority": [
                8,
                Gtk.CellRendererCombo(),
                "changed",
                super().on_cell_change,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Priority"),
                "gint",
            ],
            "requirement_code": [
                9,
                Gtk.CellRendererText(),
                "edited",
                None,
                "mvw_editing_requirement",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": False,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Code"),
                "gchararray",
            ],
            "specification": [
                10,
                Gtk.CellRendererText(),
                "edited",
                super().on_cell_edit,
                "mvw_editing_requirement",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Specification"),
                "gchararray",
            ],
            "requirement_type": [
                11,
                Gtk.CellRendererCombo(),
                "changed",
                super().on_cell_change,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Type"),
                "gchararray",
            ],
            "validated": [
                12,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Validated?"),
                "gint",
            ],
            "validated_date": [
                13,
                Gtk.CellRendererText(),
                "edited",
                None,
                "mvw_editing_requirement",
                "",
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": True,
                },
                _("Validated Date"),
                "gchararray",
            ],
            "q_clarity_0": [
                14,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Clarity Q1"),
                "gint",
            ],
            "q_clarity_1": [
                15,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Clarity Q2"),
                "gint",
            ],
            "q_clarity_2": [
                16,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Clarity Q3"),
                "gint",
            ],
            "q_clarity_3": [
                17,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Clarity Q4"),
                "gint",
            ],
            "q_clarity_4": [
                18,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Clarity Q5"),
                "gint",
            ],
            "q_clarity_5": [
                19,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Clarity Q6"),
                "gint",
            ],
            "q_clarity_6": [
                20,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Clarity Q7"),
                "gint",
            ],
            "q_clarity_7": [
                21,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Clarity Q8"),
                "gint",
            ],
            "q_clarity_8": [
                22,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Clarity Q9"),
                "gint",
            ],
            "q_complete_0": [
                23,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Complete Q1"),
                "gint",
            ],
            "q_complete_1": [
                24,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Complete Q2"),
                "gint",
            ],
            "q_complete_2": [
                25,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Complete Q3"),
                "gint",
            ],
            "q_complete_3": [
                26,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Complete Q4"),
                "gint",
            ],
            "q_complete_4": [
                27,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Complete Q5"),
                "gint",
            ],
            "q_complete_5": [
                28,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Complete Q6"),
                "gint",
            ],
            "q_complete_6": [
                29,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Complete Q7"),
                "gint",
            ],
            "q_complete_7": [
                30,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Complete Q8"),
                "gint",
            ],
            "q_complete_8": [
                31,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Complete Q9"),
                "gint",
            ],
            "q_complete_9": [
                32,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Complete Q10"),
                "gint",
            ],
            "q_consistent_0": [
                33,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Consistent Q1"),
                "gint",
            ],
            "q_consistent_1": [
                34,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Consistent Q2"),
                "gint",
            ],
            "q_consistent_2": [
                35,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Consistent Q3"),
                "gint",
            ],
            "q_consistent_3": [
                36,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Consistent Q4"),
                "gint",
            ],
            "q_consistent_4": [
                37,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Consistent Q5"),
                "gint",
            ],
            "q_consistent_5": [
                38,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Consistent Q6"),
                "gint",
            ],
            "q_consistent_6": [
                39,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Consistent Q7"),
                "gint",
            ],
            "q_consistent_7": [
                40,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Consistent Q8"),
                "gint",
            ],
            "q_consistent_8": [
                41,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Consistent Q9"),
                "gint",
            ],
            "q_verifiable_0": [
                42,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Verifiable Q1"),
                "gint",
            ],
            "q_verifiable_1": [
                43,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Verifiable Q2"),
                "gint",
            ],
            "q_verifiable_2": [
                44,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Verifiable Q3"),
                "gint",
            ],
            "q_verifiable_3": [
                45,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Verifiable Q4"),
                "gint",
            ],
            "q_verifiable_4": [
                46,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Verifiable Q5"),
                "gint",
            ],
            "q_verifiable_5": [
                47,
                Gtk.CellRendererToggle(),
                "toggled",
                None,
                "mvw_editing_requirement",
                0,
                {
                    "bg_color": "#FFFFFF",
                    "editable": True,
                    "fg_color": "#000000",
                    "visible": False,
                },
                _("Verifiable Q6"),
                "gint",
            ],
        }

        # Initialize public list class attributes.
        self.lst_owner: List[str] = [""]
        self.lst_type: List[str] = [""]

        # Initialize public scalar class attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.tvwTreeView.set_tooltip_text(
            _("Displays the hierarchical list of requirements.")
        )

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._on_module_switch, "mvwSwitchedPage")

    def _on_module_switch(self, module: str = "") -> None:
        """Respond to changes in selected Module View module (tab).

        :param module: the name of the module that was just selected.
        :return: None
        """
        _model, _row = self.tvwTreeView.selection.get_selected()

        if module == self._tag and _row is not None:
            _code = _model.get_value(
                _row, self.tvwTreeView.position["requirement_code"]
            )
            _name = _model.get_value(_row, self.tvwTreeView.position["description"])
            _title = _(f"Analyzing Requirement {_code}: {_name}")

            pub.sendMessage("request_set_title", title=_title)

    def _on_row_change(self, selection: Gtk.TreeSelection) -> None:
        """Handle events for the Requirement ModuleView RAMSTKTreeView().

        This method is called whenever a Requirement Module View
        RAMSTKTreeView() row is activated/changed.

        :param selection: the Requirement class Gtk.TreeSelection().
        :return: None
        """
        _attributes = super().on_row_change(selection)

        if _attributes:
            self._record_id = _attributes["requirement_id"]
            self._parent_id = _attributes["parent_id"]
            _attributes["owner"] = self.lst_owner.index(_attributes["owner"])
            _attributes["requirement_type"] = self.lst_type.index(
                _attributes["requirement_type"]
            )

            _title = _("Analyzing Requirement {0:s}: {1:s}").format(
                str(_attributes["requirement_code"]), str(_attributes["description"])
            )

            pub.sendMessage("selected_requirement", attributes=_attributes)
            pub.sendMessage("request_set_title", title=_title)

    def __do_load_requirement(
        self, node: treelib.Node, row: Gtk.TreeIter
    ) -> Gtk.TreeIter:
        """Load a requirement into the RAMSTKTreeView().

        :param node: the treelib Node() with the mode data to load.
        :param row: the parent row of the mode to load into the requirement
            tree.
        :return: _new_row; the row that was just populated with requirement
            data.
        :rtype: :class:`Gtk.TreeIter`
        """
        _new_row = None

        [[__, _entity]] = node.data.items()  # pylint: disable=unused-variable

        _model = self.tvwTreeView.get_model()

        _attributes = [
            _entity.revision_id,
            _entity.requirement_id,
            _entity.derived,
            _entity.description,
            _entity.figure_number,
            self.lst_owner[_entity.owner],
            _entity.page_number,
            _entity.parent_id,
            _entity.priority,
            _entity.requirement_code,
            _entity.specification,
            self.lst_type[_entity.requirement_type],
            _entity.validated,
            str(_entity.validated_date),
            _entity.q_clarity_0,
            _entity.q_clarity_1,
            _entity.q_clarity_2,
            _entity.q_clarity_3,
            _entity.q_clarity_4,
            _entity.q_clarity_5,
            _entity.q_clarity_6,
            _entity.q_clarity_7,
            _entity.q_clarity_8,
            _entity.q_complete_0,
            _entity.q_complete_1,
            _entity.q_complete_2,
            _entity.q_complete_3,
            _entity.q_complete_4,
            _entity.q_complete_5,
            _entity.q_complete_6,
            _entity.q_complete_7,
            _entity.q_complete_8,
            _entity.q_complete_9,
            _entity.q_consistent_0,
            _entity.q_consistent_1,
            _entity.q_consistent_2,
            _entity.q_consistent_3,
            _entity.q_consistent_4,
            _entity.q_consistent_5,
            _entity.q_consistent_6,
            _entity.q_consistent_7,
            _entity.q_consistent_8,
            _entity.q_verifiable_0,
            _entity.q_verifiable_1,
            _entity.q_verifiable_2,
            _entity.q_verifiable_3,
            _entity.q_verifiable_4,
            _entity.q_verifiable_5,
        ]

        try:
            _new_row = _model.append(row, _attributes)
        except (AttributeError, TypeError, ValueError):
            _new_row = None
            _message = _(
                "An error occurred when loading requirement {0} in the "
                "requirement tree.  This might indicate it was missing it's "
                "data package, some of the data in the package was missing, "
                "or some of the data was the wrong type.  Row data was: "
                "{1}"
            ).format(str(node.identifier), _attributes)
            pub.sendMessage(
                "do_log_warning_msg", logger_name="WARNING", message=_message
            )

        return _new_row


class RequirementGeneralDataPanel(RAMSTKFixedPanel):
    """Panel to display general data about the selected Requirement."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "requirement_id"
    _select_msg = "selected_requirement"
    _tag = "requirement"
    _title = _("General Requirement Information")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Requirement General Data panel."""
        super().__init__()

        # Initialize widgets.
        self.btnValidateDate: RAMSTKButton = RAMSTKButton()

        self.chkDerived: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("Requirement is derived.")
        )
        self.chkValidated: RAMSTKCheckButton = RAMSTKCheckButton(
            label=_("Requirement is validated.")
        )

        self.cmbOwner: RAMSTKComboBox = RAMSTKComboBox()
        self.cmbRequirementType: RAMSTKComboBox = RAMSTKComboBox(index=1, simple=False)
        self.cmbPriority: RAMSTKComboBox = RAMSTKComboBox()

        self.txtCode: RAMSTKEntry = RAMSTKEntry()
        self.txtFigNum: RAMSTKEntry = RAMSTKEntry()
        self.txtName: RAMSTKTextView = RAMSTKTextView(Gtk.TextBuffer())
        self.txtPageNum: RAMSTKEntry = RAMSTKEntry()
        self.txtSpecification: RAMSTKEntry = RAMSTKEntry()
        self.txtValidatedDate: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dict instance attributes.
        self.dic_attribute_widget_map = {
            "requirement_code": [
                9,
                self.txtCode,
                "changed",
                super().on_changed_entry,
                "wvw_editing_requirement",
                "",
                {
                    "width": 125,
                    "tooltip": _("A unique code for the selected requirement."),
                },
                _("Requirement Code:"),
                "gchararray",
            ],
            "description": [
                3,
                self.txtName,
                "changed",
                super().on_changed_textview,
                "wvw_editing_requirement",
                "",
                {
                    "height": 100,
                    "width": 800,
                    "tooltip": _("The description of the selected requirement."),
                },
                _("Requirement Description:"),
                "gchararray",
            ],
            "requirement_type": [
                11,
                self.cmbRequirementType,
                "changed",
                super().on_changed_combo,
                "wvw_editing_requirement",
                0,
                {
                    "tooltip": _("The type of requirement."),
                },
                _("Requirement Type:"),
                "gint",
            ],
            "derived": [
                2,
                self.chkDerived,
                "toggled",
                super().on_toggled,
                "mvw_editing_requirement",
                0,
                {
                    "tooltip": _(
                        "Indicates whether or not the selected requirement is derived."
                    ),
                    "width": 400,
                },
                "",
                "gint",
            ],
            "specification": [
                10,
                self.txtSpecification,
                "changed",
                super().on_changed_entry,
                "wvw_editing_requirement",
                "",
                {
                    "tooltip": _(
                        "The governing specification, if any, for the requirement."
                    )
                },
                _("Specification:"),
                "gchararray",
            ],
            "page_number": [
                6,
                self.txtSpecification,
                "changed",
                super().on_changed_entry,
                "wvw_editing_requirement",
                "",
                {
                    "tooltip": _(
                        "The applicable page number in the governing specification."
                    )
                },
                _("Page Number:"),
                "gchararray",
            ],
            "figure_number": [
                4,
                self.txtFigNum,
                "changed",
                super().on_changed_entry,
                "wvw_editing_requirement",
                "",
                {
                    "toolip": _(
                        "The applicable figure number in the governing specification."
                    )
                },
                _("Figure Number:"),
                "gchararray",
            ],
            "priority": [
                8,
                self.cmbPriority,
                "changed",
                super().on_changed_combo,
                "wvw_editing_requirement",
                0,
                {
                    "width": 50,
                },
                _("Priority:"),
                "gint",
            ],
            "owner": [
                5,
                self.cmbOwner,
                "changed",
                super().on_changed_combo,
                "wvw_editing_requirement",
                0,
                {
                    "tooltip": _("The organization responsible for the requirement."),
                },
                _("Owner:"),
                "gint",
            ],
            "validated": [
                12,
                self.chkValidated,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "tooltip": _(
                        "Indicates whether or not the selected requirement is "
                        "validated."
                    ),
                    "width": 400,
                },
                "",
                "gint",
            ],
            "validated_date": [
                13,
                self.txtValidatedDate,
                "changed",
                super().on_changed_entry,
                "wvw_editing_requirement",
                date.today(),
                {
                    "tooltip": _("The date the selected requirement was validated."),
                },
                _("Validated Date:"),
                "gchararray",
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel()
        super().do_set_callbacks()

        self.btnValidateDate.do_set_properties(
            height=25,
            width=25,
        )
        self.btnValidateDate.dic_handler_id["released"] = self.btnValidateDate.connect(
            "button-release-event",
            self._do_select_date,
            self.txtValidatedDate,
        )

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._do_load_code, "succeed_create_requirement_code")

    def do_load_priorities(self) -> None:
        """Load the priority RAMSTKComboBox().

        :return: None
        :rtype: None
        """
        _priorities: List[List[str]] = [["1"], ["2"], ["3"], ["4"], ["5"]]
        self.cmbPriority.do_load_combo(_priorities)

    def do_load_requirement_types(
        self, requirement_types: Dict[int, Tuple[str]]
    ) -> None:
        """Load the requirement types RAMSTKComboBox().

        :param requirement_types:
        :return: None
        :rtype: None
        """
        _requirement_types: List[Tuple[str]] = []

        # pylint: disable=unused-variable
        for __, _key in enumerate(requirement_types):
            _requirement_types.append(requirement_types[_key])
        self.cmbRequirementType.do_load_combo(entries=_requirement_types, simple=False)

    def do_load_workgroups(self, workgroups: Dict[int, Tuple[str]]) -> None:
        """Load the workgroups RAMSTKComboBox().

        :param workgroups:
        :return: None
        :rtype: None
        """
        _owners = []

        # pylint: disable=unused-variable
        for __, _key in enumerate(workgroups):
            _owners.append(workgroups[_key])
        self.cmbOwner.do_load_combo(_owners)

    def _do_load_code(self, requirement_code: int) -> None:
        """Load the Requirement code RAMSTKEntry().

        :param requirement_code: the Requirement code to load.
        :return: None
        :rtype: None
        """
        self.txtCode.do_update(str(requirement_code), signal="changed")

    @staticmethod
    def _do_select_date(
        __button: RAMSTKButton, __event: Gdk.Event, entry: RAMSTKEntry
    ) -> str:
        """Request to launch a date selection dialog.

        This method is used to select the validation date for the Requirement.

        :param __button: the ramstk.RAMSTKButton() that called this method.
        :param __event: the Gdk.Event() that called this method.
        :param entry: the Gtk.Entry() that the new date should be displayed in.
        :return: _date; the date in ISO-8601 (YYYY-mm-dd) format.
        :rtype: str
        """
        _parent = (
            entry.get_parent()
            .get_parent()
            .get_parent()
            .get_parent()
            .get_parent()
            .get_parent()
            .get_parent()
            .get_parent()
            .get_parent()
        )

        _dialog: RAMSTKDateSelect = RAMSTKDateSelect(dlgparent=_parent)

        _date = _dialog.do_run()
        _dialog.do_destroy()

        entry.do_update(_date)
        super().on_changed_entry(entry, "validated_date", "wvw_editing_requirement")

        return _date


class RequirementClarityPanel(RAMSTKFixedPanel):
    """Panel to display clarity questions about the selected Requirement."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "requirement_id"
    _select_msg = "selected_requirement"
    _tag = "requirement"
    _title = _("Clarity of Requirement")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Requirement clarity panel."""
        super().__init__()

        # Initialize widgets.
        self.chkClarityQ0: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkClarityQ1: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkClarityQ2: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkClarityQ3: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkClarityQ4: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkClarityQ5: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkClarityQ6: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkClarityQ7: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkClarityQ8: RAMSTKCheckButton = RAMSTKCheckButton()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dict instance attributes.
        self.dic_attribute_widget_map = {
            "q_clarity_0": [
                14,
                self.chkClarityQ0,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("1. The requirement clearly states what is needed or desired."),
                "gint",
            ],
            "q_clarity_1": [
                15,
                self.chkClarityQ1,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("2. The requirement is unambiguous and not open to interpretation."),
                "gint",
            ],
            "q_clarity_2": [
                16,
                self.chkClarityQ2,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "3. All terms that can have more than one meaning are qualified so "
                    "that the desired meaning is readily apparent."
                ),
                "gint",
            ],
            "q_clarity_3": [
                17,
                self.chkClarityQ3,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "4. Diagrams, drawings, etc. are used to increase understanding of "
                    "the requirement."
                ),
                "gint",
            ],
            "q_clarity_4": [
                18,
                self.chkClarityQ4,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("5. The requirement is free from spelling and grammatical errors."),
                "gint",
            ],
            "q_clarity_5": [
                19,
                self.chkClarityQ5,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "6. The requirement is written in non-technical language using the "
                    "vocabulary of the stakeholder."
                ),
                "gint",
            ],
            "q_clarity_6": [
                20,
                self.chkClarityQ6,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("7. Stakeholders understand the requirement as written."),
                "gint",
            ],
            "q_clarity_7": [
                21,
                self.chkClarityQ7,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "8. The requirement is clear enough to be turned over to an "
                    "independent group and still be understood."
                ),
                "gint",
            ],
            "q_clarity_8": [
                22,
                self.chkClarityQ8,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "9. The requirement avoids stating how the problem is "
                    "to be solved or what techniques are to be used."
                ),
                "gint",
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel(justify=Gtk.Justification.LEFT)
        super().do_set_callbacks()

        # Subscribe to PyPubSub messages.


class RequirementCompletenessPanel(RAMSTKFixedPanel):
    """Panel to display completeness questions about selected Requirement."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "requirement_id"
    _select_msg = "selected_requirement"
    _tag = "requirement"
    _title = _("Completeness of Requirement")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Requirement completeness panel."""
        super().__init__()

        # Initialize widgets.
        self.chkCompleteQ0: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkCompleteQ1: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkCompleteQ2: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkCompleteQ3: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkCompleteQ4: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkCompleteQ5: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkCompleteQ6: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkCompleteQ7: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkCompleteQ8: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkCompleteQ9: RAMSTKCheckButton = RAMSTKCheckButton()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dict instance attributes.
        self.dic_attribute_widget_map = {
            "q_complete_0": [
                23,
                self.chkCompleteQ0,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "1. Performance objectives are properly documented from the "
                    "user's point of view."
                ),
                "gint",
            ],
            "q_complete_1": [
                24,
                self.chkCompleteQ1,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("2. No necessary information is missing from the requirement."),
                "gint",
            ],
            "q_complete_2": [
                25,
                self.chkCompleteQ2,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("3. The requirement has been assigned a priority."),
                "gint",
            ],
            "q_complete_3": [
                26,
                self.chkCompleteQ3,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "4. The requirement is realistic given the technology that will "
                    "be used to implement the system."
                ),
                "gint",
            ],
            "q_complete_4": [
                27,
                self.chkCompleteQ4,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "5. The requirement is feasible to implement given the defined "
                    "project time frame, scope, structure and budget."
                ),
                "gint",
            ],
            "q_complete_5": [
                28,
                self.chkCompleteQ5,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "6. If the requirement describes something as a 'standard' the "
                    "specific source is cited."
                ),
                "gint",
            ],
            "q_complete_6": [
                29,
                self.chkCompleteQ6,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("7. The requirement is relevant to the problem and its solution."),
                "gint",
            ],
            "q_complete_7": [
                30,
                self.chkCompleteQ7,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("8. The requirement contains no implied design details."),
                "gint",
            ],
            "q_complete_8": [
                31,
                self.chkCompleteQ8,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "9. The requirement contains no implied implementation "
                    "constraints."
                ),
                "gint",
            ],
            "q_complete_9": [
                32,
                self.chkCompleteQ9,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "10. The requirement contains no implied project management "
                    "constraints."
                ),
                "gint",
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel(justify=Gtk.Justification.LEFT)
        super().do_set_callbacks()

        # Subscribe to PyPubSub messages.


class RequirementConsistencyPanel(RAMSTKFixedPanel):
    """Panel to display consistency questions about selected Requirement."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "requirement_id"
    _select_msg = "selected_requirement"
    _tag = "requirement"
    _title = _("Consistency of Requirement")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Requirement consistency panel."""
        super().__init__()

        # Initialize widgets.
        self.chkConsistentQ0: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkConsistentQ1: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkConsistentQ2: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkConsistentQ3: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkConsistentQ4: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkConsistentQ5: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkConsistentQ6: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkConsistentQ7: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkConsistentQ8: RAMSTKCheckButton = RAMSTKCheckButton()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dict instance attributes.
        self.dic_attribute_widget_map = {
            "q_consistent_0": [
                33,
                self.chkConsistentQ0,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "1. The requirement describes a single need or want; it could not "
                    "be broken into several different requirements."
                ),
                "gint",
            ],
            "q_consistent_1": [
                34,
                self.chkConsistentQ1,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "2. The requirement requires non-standard hardware or must use "
                    "software to implement."
                ),
                "gint",
            ],
            "q_consistent_2": [
                35,
                self.chkConsistentQ2,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("3. The requirement can be implemented within known constraints."),
                "gint",
            ],
            "q_consistent_3": [
                36,
                self.chkConsistentQ3,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "4. The requirement provides an adequate basis for design and "
                    "testing."
                ),
                "gint",
            ],
            "q_consistent_4": [
                37,
                self.chkConsistentQ4,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "5. The requirement adequately supports the business goal of the "
                    "project."
                ),
                "gint",
            ],
            "q_consistent_5": [
                38,
                self.chkConsistentQ5,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "6. The requirement does not conflict with some constraint, policy "
                    "or regulation."
                ),
                "gint",
            ],
            "q_consistent_6": [
                39,
                self.chkConsistentQ6,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("7. The requirement does not conflict with another requirement."),
                "gint",
            ],
            "q_consistent_7": [
                40,
                self.chkConsistentQ7,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("8. The requirement is not a duplicate of another requirement."),
                "gint",
            ],
            "q_consistent_8": [
                41,
                self.chkConsistentQ8,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _("9. The requirement is in scope for the project."),
                "gint",
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel(justify=Gtk.Justification.LEFT)
        super().do_set_callbacks()

        # Subscribe to PyPubSub messages.


class RequirementVerifiabilityPanel(RAMSTKFixedPanel):
    """Panel to display verifiability questions about selected Requirement."""

    # Define private dictionary class attributes.

    # Define private list class attributes.

    # Define private scalar class attributes.
    _record_field = "requirement_id"
    _select_msg = "selected_requirement"
    _tag = "requirement"
    _title = _("Verifiability of Requirement")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize an instance of the Requirement verifiability panel."""
        super().__init__()

        # Initialize widgets.
        self.chkVerifiableQ0: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkVerifiableQ1: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkVerifiableQ2: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkVerifiableQ3: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkVerifiableQ4: RAMSTKCheckButton = RAMSTKCheckButton()
        self.chkVerifiableQ5: RAMSTKCheckButton = RAMSTKCheckButton()

        # Initialize private dict instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dict instance attributes.
        self.dic_attribute_widget_map = {
            "q_verifiable_0": [
                42,
                self.chkVerifiableQ0,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "1. The requirement is verifiable by testing, demonstration, "
                    "review, or analysis."
                ),
                "gint",
            ],
            "q_verifiable_1": [
                43,
                self.chkVerifiableQ1,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "2. The requirement lacks 'weasel words' (e.g. various, mostly, "
                    "suitable, integrate, maybe, consistent, robust, modular, "
                    "user-friendly, superb, good)."
                ),
                "gint",
            ],
            "q_verifiable_2": [
                44,
                self.chkVerifiableQ2,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "3. Any performance criteria are quantified such that they are "
                    "testable."
                ),
                "gint",
            ],
            "q_verifiable_3": [
                45,
                self.chkVerifiableQ3,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "4. Independent testing would be able to determine whether the "
                    "requirement has been satisfied."
                ),
                "gint",
            ],
            "q_verifiable_4": [
                46,
                self.chkVerifiableQ4,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "5. The task(s) that will validate and verify the final design "
                    "satisfies the requirement have been identified."
                ),
                "gint",
            ],
            "q_verifiable_5": [
                47,
                self.chkVerifiableQ5,
                "toggled",
                super().on_toggled,
                "wvw_editing_requirement",
                0,
                {
                    "height": 30,
                },
                _(
                    "6. The identified V&amp;V task(s) have been added to the "
                    "validation plan (e.g., DVP)"
                ),
                "gint",
            ],
        }

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

        super().do_set_properties()
        super().do_make_panel(justify=Gtk.Justification.LEFT)
        super().do_set_callbacks()

        # Subscribe to PyPubSub messages.



# Standard Library Imports
from typing import Any

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE as RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord as RAMSTKBaseRecord

class RAMSTKProgramInfoRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    __defaults__: Any
    __tablename__: str
    __table_args__: Any
    revision_id: Any
    function_active: Any
    requirement_active: Any
    hardware_active: Any
    software_active: Any
    rcm_active: Any
    testing_active: Any
    incident_active: Any
    survival_active: Any
    vandv_active: Any
    hazard_active: Any
    stakeholder_active: Any
    allocation_active: Any
    similar_item_active: Any
    fmea_active: Any
    pof_active: Any
    rbd_active: Any
    fta_active: Any
    created_on: Any
    created_by: Any
    last_saved: Any
    last_saved_by: Any
    def get_attributes(self): ...



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       ramstk.__main__.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""The main program for the RAMSTK application."""

# Standard Library Imports
import os
import shutil
import sys
from time import sleep
from typing import Tuple

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKSiteConfiguration, RAMSTKUserConfiguration
from ramstk.db import BaseDatabase
from ramstk.exim import Export, Import
from ramstk.logger import RAMSTKLogManager
from ramstk.models import (
    RAMSTKActionTable,
    RAMSTKAllocationTable,
    RAMSTKCauseTable,
    RAMSTKCommonDB,
    RAMSTKControlTable,
    RAMSTKDesignElectricTable,
    RAMSTKDesignMechanicTable,
    RAMSTKEnvironmentTable,
    RAMSTKFailureDefinitionTable,
    RAMSTKFMEAView,
    RAMSTKFunctionTable,
    RAMSTKHardwareBoMView,
    RAMSTKHardwareTable,
    RAMSTKHazardTable,
    RAMSTKMechanismTable,
    RAMSTKMILHDBK217FTable,
    RAMSTKMissionPhaseTable,
    RAMSTKMissionTable,
    RAMSTKModeTable,
    RAMSTKNSWCTable,
    RAMSTKOpLoadTable,
    RAMSTKOpStressTable,
    RAMSTKPoFView,
    RAMSTKProgramDB,
    RAMSTKProgramInfoTable,
    RAMSTKProgramStatusTable,
    RAMSTKReliabilityTable,
    RAMSTKRequirementTable,
    RAMSTKRevisionTable,
    RAMSTKSimilarItemTable,
    RAMSTKSiteInfoTable,
    RAMSTKStakeholderTable,
    RAMSTKTestMethodTable,
    RAMSTKUsageProfileView,
    RAMSTKValidationTable,
)
from ramstk.utilities import file_exists
from ramstk.views.gtk3 import Gtk, RAMSTKDesktop, _
from ramstk.views.gtk3.widgets import RAMSTKDatabaseSelect


def do_connect_to_site_db(conn_info) -> BaseDatabase:
    """Connect to the site (common) database.

    :param conn_info: the site database connection information.
    :return: _site_db
    :rtype: BaseDatabase
    """
    pub.sendMessage(
        "do_log_info_msg",
        logger_name="INFO",
        message="Connecting to the RAMSTK common database {0} on {1} "
        "using port {2}.".format(
            conn_info["database"], conn_info["host"], conn_info["port"]
        ),
    )

    _site_db = BaseDatabase()
    _site_db.do_connect(conn_info)
    pub.sendMessage(
        "do_log_info_msg",
        logger_name="INFO",
        message="Connected to the RAMSTK common database {0:s}.".format(
            conn_info["database"]
        ),
    )

    return _site_db


def do_first_run(configuration: RAMSTKSiteConfiguration) -> None:
    """Raise dialog to setup site database.

    :param configuration: the RAMSTKSiteConfiguration() instance.
    :return: None
    :rtype: None
    """
    _dialog = RAMSTKDatabaseSelect(
        dlgtitle=_("Set up RAMSTK Site Database Server Connection"),
        dao=BaseDatabase(),
        database=configuration.RAMSTK_COM_INFO,
        icons={
            "refresh": configuration.RAMSTK_SITE_DIR + "/icons/32x32/view-refresh.png",
            "save": configuration.RAMSTK_SITE_DIR + "/icons/32x32/save.png",
        },
    )

    if _dialog.do_run() == Gtk.ResponseType.OK:
        _site_dir = configuration.RAMSTK_SITE_DIR
        _home = os.path.expanduser("~")
        _user_dir = _home + "/.config/RAMSTK"
        if not os.path.isdir(_user_dir):
            shutil.copytree(_site_dir + "/icons", _user_dir + "/icons/")
            shutil.copytree(_site_dir + "/layouts", _user_dir + "/layouts/")
            shutil.copy(_site_dir + "/RAMSTK.toml", _user_dir)
            shutil.copy(_site_dir + "/postgres_program_db.sql", _user_dir)
            os.makedirs(_user_dir + "/logs")

        configuration.RAMSTK_COM_INFO = _dialog.database
    else:
        sys.exit(0)

    _dialog.do_destroy()


def do_initialize_databases(
    configuration: RAMSTKUserConfiguration, site_db: BaseDatabase
) -> Tuple[RAMSTKProgramDB, RAMSTKCommonDB]:
    """Initialize the databases for the current instance of RAMSTK.

    :param configuration: the instance of the user configuration object to associate
        with this database model.
    :param site_db: the instance of the site data access object to associate with
        this database model.
    :return: _program_db, _site_db
    :rtype: RAMSTKProgramDB, RAMSTKCommonDB
    """
    _program_db = RAMSTKProgramDB()
    _site_db = RAMSTKCommonDB()

    _program_db.dic_tables["action"] = RAMSTKActionTable()
    _program_db.dic_tables["allocation"] = RAMSTKAllocationTable()
    _program_db.dic_tables["cause"] = RAMSTKCauseTable()
    _program_db.dic_tables["control"] = RAMSTKControlTable()
    _program_db.dic_tables["design_electric"] = RAMSTKDesignElectricTable()
    _program_db.dic_tables["design_mechanic"] = RAMSTKDesignMechanicTable()
    _program_db.dic_tables["environment"] = RAMSTKEnvironmentTable()
    _program_db.dic_tables["failure_definition"] = RAMSTKFailureDefinitionTable()
    _program_db.dic_tables["function"] = RAMSTKFunctionTable()
    _program_db.dic_tables["hardware"] = RAMSTKHardwareTable()
    _program_db.dic_tables["hazards"] = RAMSTKHazardTable()
    _program_db.dic_tables["mechanism"] = RAMSTKMechanismTable()
    _program_db.dic_tables["milhdbk217f"] = RAMSTKMILHDBK217FTable()
    _program_db.dic_tables["mission"] = RAMSTKMissionTable()
    _program_db.dic_tables["mission_phase"] = RAMSTKMissionPhaseTable()
    _program_db.dic_tables["mode"] = RAMSTKModeTable()
    _program_db.dic_tables["nswc"] = RAMSTKNSWCTable()
    _program_db.dic_tables["opload"] = RAMSTKOpLoadTable()
    _program_db.dic_tables["opstress"] = RAMSTKOpStressTable()
    _program_db.dic_tables["program_info"] = RAMSTKProgramInfoTable()
    _program_db.dic_tables["program_status"] = RAMSTKProgramStatusTable()
    _program_db.dic_tables["reliability"] = RAMSTKReliabilityTable()
    _program_db.dic_tables["requirement"] = RAMSTKRequirementTable()
    _program_db.dic_tables["revision"] = RAMSTKRevisionTable()
    _program_db.dic_tables["similar_item"] = RAMSTKSimilarItemTable()
    _program_db.dic_tables["stakeholder"] = RAMSTKStakeholderTable()
    _program_db.dic_tables["test_method"] = RAMSTKTestMethodTable()
    _program_db.dic_tables["validation"] = RAMSTKValidationTable()
    _program_db.dic_tables["export"] = Export()
    _program_db.dic_tables["import"] = Import()
    _program_db.user_configuration = configuration

    _program_db.dic_views["fmea"] = RAMSTKFMEAView()
    _program_db.dic_views["hardwarebom"] = RAMSTKHardwareBoMView(
        hr_multiplier=configuration.RAMSTK_HR_MULTIPLIER
    )
    _program_db.dic_views["pof"] = RAMSTKPoFView()
    _program_db.dic_views["usage_profile"] = RAMSTKUsageProfileView()

    # noinspection PyTypeChecker
    _site_db.common_dao = site_db
    _site_db.dic_tables["options"] = RAMSTKSiteInfoTable()
    _site_db.dic_tables["options"].dao = site_db
    _site_db.dic_tables["options"].do_select_all({"site_id": 1})

    return _program_db, _site_db


def do_initialize_loggers(log_file: str, log_level: str) -> RAMSTKLogManager:
    """Initialize the loggers for the current instance of RAMSTK.

    :param log_file: the absolute path to the file used to capture logs.
    :param log_level: the lowest level of messages to log as defined by the
        user's configuration.
    :return: _logger; the RAMSTKLogManager() managing the loggers.
    :rtype: :class:`RAMSTKLogManager`
    """
    if file_exists(log_file):
        os.remove(log_file)

    _logger: RAMSTKLogManager = RAMSTKLogManager(log_file)

    for _level in ["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]:
        _logger.do_create_logger(
            _level,
            log_level,
            to_tty={
                "DEBUG": False,
                "INFO": False,
                "WARNING": False,
                "ERROR": True,
                "CRITICAL": True,
            }[_level],
        )

    return _logger


def do_read_site_configuration() -> RAMSTKSiteConfiguration:
    """Create a site configuration instance.

    :return: _configuration; the RAMSTKSiteConfiguraion() instance to use for
        this run of RAMSTK.
    :rtype: :class:`ramstk.configuration.RAMSTKSiteConfiguration`
    """

    def on_fail_create_site_configuration(error_message: str) -> None:
        """Log error message when there's a failure to create the site conf.

        :param error_message: the error message raised by the failure.
        :return: None
        :rtype: None
        """
        pub.sendMessage("do_log_debug_msg", logger_name="DEBUG", message=error_message)

    pub.subscribe(on_fail_create_site_configuration, "fail_create_site_configuration")

    pub.sendMessage(
        "do_log_info_msg",
        logger_name="INFO",
        message="Reading the site configuration file.",
    )

    _configuration = RAMSTKSiteConfiguration()
    _configuration.set_site_directories()
    _configuration.get_site_configuration()

    if _configuration.RAMSTK_COM_INFO["user"] == "first_run":
        do_first_run(_configuration)
        _configuration.set_site_configuration()

    pub.sendMessage(
        "do_log_info_msg",
        logger_name="INFO",
        message="Read the site configuration file.",
    )

    return _configuration


def do_read_user_configuration() -> Tuple[RAMSTKUserConfiguration, RAMSTKLogManager]:
    """Create a user configuration instance.

    :return: _configuration; the RAMSTKUserConfiguraion() instance to use for
        this run of RAMSTK.
    :rtype: :class:`ramstk.configuration.RAMSTKUserConfiguration`
    """

    def on_fail_create_user_configuration(error_message: str) -> None:
        """Log error message when there's a failure to create the user conf.

        :param error_message: the error message raised by the failure.
        :return: None
        :rtype: None
        """
        print(error_message)

    pub.subscribe(on_fail_create_user_configuration, "fail_create_user_configuration")

    _configuration = RAMSTKUserConfiguration()

    _configuration.set_user_directories()
    _configuration.get_user_configuration()

    if _configuration.RAMSTK_DATA_DIR == "":
        _configuration.RAMSTK_DATA_DIR = _configuration.RAMSTK_CONF_DIR + "/layouts"
        _configuration.set_user_configuration()

    if _configuration.RAMSTK_ICON_DIR == "":
        _configuration.RAMSTK_ICON_DIR = _configuration.RAMSTK_CONF_DIR + "/icons"
        _configuration.set_user_configuration()

    if _configuration.RAMSTK_LOG_DIR == "":
        _configuration.RAMSTK_LOG_DIR = _configuration.RAMSTK_CONF_DIR + "/logs"
        _configuration.set_user_configuration()

    _logger = do_initialize_loggers(
        _configuration.RAMSTK_USER_LOG, _configuration.RAMSTK_LOGLEVEL
    )

    return _configuration, _logger


def the_one_ring() -> None:
    """Execute the main function for RAMSTK."""
    # ISSUE: Implement splash screen.
    # //
    # // Add a splash screen to the launch of RAMSTK.
    # //
    # // labels: globalbacklog, normal
    # splScreen = SplashScreen()
    # If you don't do this, the splash screen will show, but won't render it's
    # contents
    # while Gtk.events_pending():
    #     Gtk.main_iteration()

    site_configuration = do_read_site_configuration()

    # Read the user configuration file and create a logger.  The user
    # configuration file contains information needed to create the logger so
    # it must come first.
    user_configuration, _logger = do_read_user_configuration()

    pub.sendMessage(
        "do_log_debug_msg",
        logger_name="DEBUG",
        message="Validating the RAMSTK license.",
    )
    pub.sendMessage(
        "do_log_debug_msg", logger_name="DEBUG", message="Validated the RAMSTK license."
    )

    site_db = do_connect_to_site_db(site_configuration.RAMSTK_COM_INFO)

    pub.sendMessage(
        "do_log_info_msg",
        logger_name="INFO",
        message="Initializing the RAMSTK application.",
    )
    print(site_db)
    _program_db, _site_db = do_initialize_databases(  # pylint: disable=unused-variable
        user_configuration, site_db
    )
    print(_site_db, _site_db.common_dao)
    user_configuration = _site_db.do_load_site_variables(user_configuration)

    pub.sendMessage(
        "do_log_info_msg",
        logger_name="INFO",
        message="Initialized the RAMSTK application.",
    )
    pub.sendMessage(
        "do_log_info_msg", logger_name="INFO", message="Launching RAMSTK GUI."
    )

    # Create the RAMSTK Book.  This needs to be initialized after reading the
    # configuration and creating the logger.
    RAMSTKDesktop([user_configuration, site_configuration], _logger)

    pub.sendMessage(
        "do_log_info_msg",
        logger_name="INFO",
        message="Launched RAMSTK GUI.",
    )

    sleep(1)
    # splScreen.window.destroy()

    Gtk.main()



# Standard Library Imports
from typing import Any, Union

# RAMSTK Package Imports
from ramstk.views.gtk3 import GdkPixbuf as GdkPixbuf
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _

# RAMSTK Local Imports
from .widget import RAMSTKWidget as RAMSTKWidget

def do_make_buttonbox(
    view: Any, **kwargs: Any
) -> Union[Gtk.HButtonBox, Gtk.VButtonBox]: ...

class RAMSTKButton(Gtk.Button, RAMSTKWidget):
    _default_height: int = ...
    _default_width: int = ...
    def __init__(self, label: str = ...) -> None: ...
    def do_set_properties(self, **kwargs: Any) -> None: ...

class RAMSTKCheckButton(Gtk.CheckButton, RAMSTKWidget):
    _default_height: int = ...
    _default_width: int = ...
    def __init__(self, label: str = ...) -> None: ...
    def do_set_properties(self, **kwargs: Any) -> None: ...
    def do_update(self, value: int, signal: str = ...) -> None: ...

class RAMSTKFileChooserButton(Gtk.FileChooserButton, RAMSTKWidget):
    def __init__(self, label: str = "...") -> None: ...
    def do_set_properties(self, **kwargs: Any) -> None: ...

class RAMSTKOptionButton(Gtk.RadioButton, RAMSTKWidget):
    def __init__(self, group: Gtk.RadioButton = ..., label: str = ...) -> None: ...

class RAMSTKSpinButton(Gtk.SpinButton, RAMSTKWidget):
    _default_height: int = ...
    _default_width: int = ...
    def __init__(self) -> None: ...
    def do_get_text(self) -> float: ...
    def do_set_properties(self, **kwargs: Any) -> None: ...
    def do_update(self, value: int, signal: str = ...) -> None: ...



# Standard Library Imports
from typing import Any, Callable, List, Tuple

# Third Party Imports
import numpy as np

def do_calculate_beta_bounds(
    minimum: float, likely: float, maximum: float, alpha: float
) -> Tuple[float, float, float, float]: ...
def do_calculate_fisher_information(
    model: Callable, p0: List[float], data: np.ndarray, noise: Any = ...
) -> np.ndarray: ...



# pylint: disable=non-parent-init-called
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.widgets.widget.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2020 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK GTK3 Base Widget Module."""

# Standard Library Imports
from typing import Any, Dict

# RAMSTK Package Imports
from ramstk.views.gtk3 import GObject, _


class RAMSTKWidget:
    """The RAMSTK Base Widget class."""

    # Define private scalar class attributes.
    _default_height = -1
    _default_width = -1

    def __init__(self) -> None:
        """Create RAMSTK Base widgets."""
        GObject.GObject.__init__(self)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.
        self.dic_handler_id: Dict[str, int] = {"": 0}

        # Initialize public list attributes.

        # Initialize public scalar attributes.
        self.height: int = -1
        self.width: int = -1

    def do_set_properties(self, **kwargs: Any) -> None:
        """Set the properties of the RAMSTK combobox.

        :param **kwargs: See below

        :Keyword Arguments:
            * *height* (int) -- height of the RAMSTKWidget().
            * *tooltip* (str) -- the tooltip, if any, for the combobox.
                Default is a message to file a QA-type issue to have one added.
            * *width* (int) -- width of the RAMSTKWidget().
        :return: None
        :rtype: None
        """
        _can_focus = kwargs.get("can_focus", True)
        _height = kwargs.get("height", self._default_height)
        _tooltip = kwargs.get(
            "tooltip",
            _("Missing tooltip, please file a quality type issue to have one added."),
        )
        _width = kwargs.get("width", self._default_width)

        if _height == 0:
            _height = self._default_height
        if _width == 0:
            _width = self._default_width

        self.height = _height
        self.width = _width
        self.set_property("can-focus", _can_focus)  # type: ignore
        self.set_property("height-request", _height)  # type: ignore
        self.set_property("tooltip-markup", _tooltip)  # type: ignore
        self.set_property("width-request", _width)  # type: ignore



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.hardware.components.miscellaneous.py is part of the RAMSTK
#       Project.
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Miscellaneous Results Panel."""

# Standard Library Imports
from typing import Any, Dict, List

# RAMSTK Package Imports
from ramstk.views.gtk3 import _
from ramstk.views.gtk3.milhdbk217f import MilHdbk217FResultPanel
from ramstk.views.gtk3.widgets import RAMSTKEntry


class MiscellaneousMilHdbk217FResultPanel(MilHdbk217FResultPanel):
    """Display Misc assessment results attribute data in the RAMSTK Work Book.

    The Miscellaneous hardware item assessment result view displays all the
    assessment results for the selected miscellaneous hardware item.  This
    includes, currently, results for MIL-HDBK-217FN2 parts count and
    MIL-HDBK-217FN2 part stress methods.  The attributes of a miscellaneous
    hardware item assessment result view are:

    :ivar txtPiA: displays the application factor for the miscellaneous
        hardware item.
    :ivar txtPiU: displays the utilization factor for the miscellaneous
        hardware item.
    """

    # Define private dict class attributes.
    _dic_part_stress: Dict[int, str] = {
        1: '<span foreground="blue">\u03BB<sub>p</sub> = '
        "\u03BB<sub>b</sub>\u03C0<sub>Q</sub>\u03C0<sub>E</sub></span>",
        2: '<span foreground="blue">\u03BB<sub>p</sub> = '
        "\u03BB<sub>b</sub>\u03C0<sub>Q</sub>\u03C0<sub>E</sub></span>",
        3: '<span foreground="blue">\u03BB<sub>p</sub> = '
        "\u03BB<sub>b</sub>\u03C0<sub>E</sub></span>",
        4: '<span foreground="blue">\u03BB<sub>p</sub> = '
        "\u03BB<sub>b</sub>\u03C0<sub>U</sub>\u03C0<sub>A</sub>\u03C0<sub>E"
        "</sub></span> ",
    }

    # Define private class list class attributes.

    # Define private scalar class attributes.
    _title: str = _("Miscellaneous Component MIL-HDBK-217F Results")

    # Define public dictionary class attributes.

    # Define public list class attributes.

    # Define public scalar class attributes.

    def __init__(self) -> None:
        """Initialize instance of the Miscellaneous assessment result view."""
        super().__init__()

        # Initialize widgets.
        self.txtPiA: RAMSTKEntry = RAMSTKEntry()
        self.txtPiU: RAMSTKEntry = RAMSTKEntry()

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.
        self.dic_attribute_widget_map: Dict[str, List[Any]] = {
            "hazard_rate_model": [
                13,
                self.lblModel,
                "",
                None,
                "",
                "",
                {
                    "tooltip": _(
                        "The assessment model used to calculate the hazard rate."
                    ),
                },
                "",
            ],
            "lambda_b": [
                23,
                self.txtLambdaB,
                "",
                None,
                "",
                0.0,
                {
                    "tooltip": _("The base hazard rate."),
                },
                "\u03BB<sub>b</sub>:",
            ],
            "pi_q": [
                30,
                self.txtPiQ,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The quality factor."),
                },
                "\u03C0<sub>Q</sub>:",
            ],
            "pi_e": [
                19,
                self.txtPiE,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The environment factor."),
                },
                "\u03C0<sub>E</sub>:",
            ],
            "pi_u": [
                35,
                self.txtPiU,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The utilization factor."),
                },
                "\u03C0<sub>U</sub>:",
            ],
            "pi_a": [
                12,
                self.txtPiA,
                "",
                None,
                "",
                1.0,
                {
                    "tooltip": _("The application factor."),
                },
                "\u03C0<sub>A</sub>:",
            ],
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        super().do_set_properties()
        super().do_make_panel()

        # Subscribe to PyPubSub messages.

    def _do_load_entries(self, attributes: Dict[str, Any]) -> None:
        """Load the miscellaneous devices assessment results page.

        :param attributes: the attributes dictionary for the selected
            Miscellaneous item.
        :return: None
        :rtype: None
        """
        super().do_load_entries(attributes)

        self.txtPiU.set_sensitive(False)
        self.txtPiA.set_sensitive(False)

        if self.category_id == 10 and self._hazard_rate_method_id == 2:
            self.txtPiU.do_update(str(self.fmt.format(attributes["piU"])))
            self.txtPiA.do_update(str(self.fmt.format(attributes["piA"])))



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.books.listbook.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK GTK3 List Book Module."""

# Standard Library Imports
from typing import Dict, List

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3.failure_definition import FailureDefinitionListView
from ramstk.views.gtk3.stakeholder import StakeholderListView
from ramstk.views.gtk3.usage_profile import UsageProfileListView
from ramstk.views.gtk3.widgets import RAMSTKBaseBook


class RAMSTKListBook(RAMSTKBaseBook):
    """The List Book class for the GTK3 multiple window interface.

    The List Book provides the container for any List Views and Matrix Views
    associated with the RAMSTK module selected in the RAMSTK Module View.
    Attributes of the List Book are:

    :ivar dict dic_list_view: dictionary containing the List Views and/or
        Matrix Views to load into the RAMSTK List Book for each RAMSTK module.
        Key is the RAMSTK module name; value is a list of Views associated with
        that RAMSTK module.
    """

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize an instance of the RAMSTK List View class.

        :param configuration: the RAMSTKUserConfiguration() class instance.
        :param logger: the RAMSTKLogManager class instance.
        """
        RAMSTKBaseBook.__init__(self, configuration)

        # Initialize private dictionary attributes.
        self._dic_list_views: Dict[str, List[object]] = {
            "revision": [
                UsageProfileListView(configuration, logger),
                FailureDefinitionListView(configuration, logger),
            ],
            "function": [],
            "requirement": [
                StakeholderListView(configuration, logger),
            ],
            "hardware": [],
            "validation": [],
        }

        # Initialize private list attributes.

        # Initialize private scalar attributes.
        self._module: str = ""

        # Initialize public dictionary attributes.

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self._set_properties("listbook")

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._on_module_change, "mvwSwitchedPage")
        pub.subscribe(self._on_close, "succeed_closed_program")

    def _on_close(self) -> None:
        """Clear the List Views when a RAMSTK Program database is closed.

        :return: None
        :rtype: None
        """
        for _listviews in self._dic_list_views.items():
            for _listview in _listviews[1]:
                # pylint: disable=protected-access
                _treeview = _listview._pnlPanel.tvwTreeView
                _model = _treeview.get_model()
                _columns = _treeview.get_columns()
                for _column in _columns:
                    _treeview.remove_column(_column)

                _model.clear()

    def _on_module_change(self, module: str = "") -> None:
        """Load List Views for the RAMSTK module selected in the Module Book.

        :param module: the name of the RAMSTK workflow module that has
            been selected in the Module Book.
        :return: None
        :rtype: None
        """
        self._module = module

        for _page in self.get_children():
            self.remove(_page)

        try:
            for _list in self._dic_list_views[module]:
                self.insert_page(_list, _list.hbx_tab_label, -1)
        except KeyError:
            pass



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.mission_phase.record.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Mission Phase Record Model."""

# Third Party Imports
from sqlalchemy import Column, Float, ForeignKeyConstraint, Integer, String
from sqlalchemy.orm import relationship

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKMissionPhaseRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent ramstk_mission_phase in the RAMSTK Program database.

    This table shares a Many-to-One relationship with ramstk_mission. This table
    shares a One-to-Many relationship with ramstk_environment.
    """

    __defaults__ = {"description": "", "name": "", "phase_start": 0.0, "phase_end": 0.0}
    __tablename__ = "ramstk_mission_phase"
    __table_args__ = (
        ForeignKeyConstraint(
            ["fld_revision_id", "fld_mission_id"],
            [
                "ramstk_mission.fld_revision_id",
                "ramstk_mission.fld_mission_id",
            ],
        ),
        {"extend_existing": True},
    )

    revision_id = Column(
        "fld_revision_id",
        Integer,
        primary_key=True,
        nullable=False,
    )
    mission_id = Column(
        "fld_mission_id",
        Integer,
        primary_key=True,
        nullable=False,
    )
    phase_id = Column(
        "fld_phase_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )

    description = Column("fld_description", String, default=__defaults__["description"])
    name = Column("fld_name", String(256), default=__defaults__["name"])
    phase_start = Column("fld_phase_start", Float, default=__defaults__["phase_start"])
    phase_end = Column("fld_phase_end", Float, default=__defaults__["phase_end"])

    # Define the relationships to other tables in the RAMSTK Program database.
    mission: relationship = relationship(
        "RAMSTKMissionRecord",
        back_populates="phase",
    )
    environment: relationship = relationship(
        "RAMSTKEnvironmentRecord",
        back_populates="phase",
        cascade="all,delete",
    )

    is_mission = False
    is_phase = True
    is_env = False

    def get_attributes(self):
        """Retrieve current values of the Mission Phase data model attributes.

        :return: value of instance attributes
        :rtype: tuple
        """
        _values = {
            "mission_id": self.mission_id,
            "phase_id": self.phase_id,
            "description": self.description,
            "name": self.name,
            "phase_start": self.phase_start,
            "phase_end": self.phase_end,
        }

        return _values



# type: ignore
# -*- coding: utf-8 -*-
#
#       ramstk.analyses.prediction.Meter.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Meter MIL-HDBK-217F Calculations Module."""

# Standard Library Imports
from typing import Any, Dict

PART_COUNT_LAMBDA_B = {
    1: {
        1: [
            10.0,
            20.0,
            120.0,
            70.0,
            180.0,
            50.0,
            80.0,
            160.0,
            250.0,
            260.0,
            5.0,
            140.0,
            380.0,
            0.0,
        ],
        2: [
            15.0,
            30.0,
            180.0,
            105.0,
            270.0,
            75.0,
            120.0,
            240.0,
            375.0,
            390.0,
            7.5,
            210.0,
            570.0,
            0.0,
        ],
        3: [
            40.0,
            80.0,
            480.0,
            280.0,
            720.0,
            200.0,
            320.0,
            640.0,
            1000.0,
            1040.0,
            20.0,
            560.0,
            1520.0,
            0.0,
        ],
    },
    2: {
        1: [0.09, 0.36, 2.3, 1.1, 3.2, 2.5, 3.8, 5.2, 6.6, 5.4, 0.099, 5.4, 0.0, 0.0],
        2: [0.15, 0.61, 2.8, 1.8, 5.4, 4.3, 6.4, 8.9, 11.0, 9.2, 0.17, 9.2, 0.0, 0.0],
    },
}
PART_COUNT_PI_Q = {1: [1.0, 1.0], 2: [1.0, 3.4]}
PART_STRESS_LAMBDA_B = {1: [20.0, 30.0, 80.0], 2: 0.09}
PART_STRESS_PI_Q = {2: [1.0, 3.4]}
PI_E = {
    1: [1.0, 2.0, 12.0, 7.0, 18.0, 5.0, 8.0, 16.0, 25.0, 26.0, 0.5, 14.0, 38.0, 0.0],
    2: [1.0, 4.0, 25.0, 12.0, 35.0, 28.0, 42.0, 58.0, 73.0, 60.0, 1.1, 60.0, 0.0, 0.0],
}
PI_F = [1.0, 1.0, 2.8]


def calculate_part_count(**attributes: Dict[str, Any]) -> float:
    """Wrap get_part_count_lambda_b().

    This wrapper allows us to pass an attributes dict from a generic parts
    count function.

    :param attributes: the attributes for the connection being calculated.
    :return: _base_hr; the parts count base hazard rates.
    :rtype: float
    """
    return get_part_count_lambda_b(
        subcategory_id=attributes["subcategory_id"],
        type_id=attributes["type_id"],
        environment_active_id=attributes["environment_active_id"],
    )


def calculate_part_stress(**attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the part stress hazard rate for a meter.

    This function calculates the MIL-HDBK-217F hazard rate using the part
    stress method.

    :return: attributes; the keyword argument (hardware attribute)
        dictionary with updated values.
    :rtype: dict
    """
    attributes["lambda_b"] = get_part_stress_lambda_b(
        attributes["subcategory_id"], attributes["type_id"]
    )
    attributes["piT"] = get_temperature_stress_factor(
        attributes["temperature_active"], attributes["temperature_rated_max"]
    )

    # Determine the application factor (piA) and function factor (piF).
    if attributes["subcategory_id"] == 2:
        attributes["piA"] = 1.7 if (attributes["type_id"]) - (1) else 1.0
        attributes["piF"] = PI_F[attributes["application_id"] - 1]

    attributes["hazard_rate_active"] = attributes["lambda_b"] * attributes["piE"]
    if attributes["subcategory_id"] == 2:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"]
            * attributes["piA"]
            * attributes["piF"]
            * attributes["piQ"]
        )
    elif attributes["subcategory_id"] == 1:
        attributes["hazard_rate_active"] = (
            attributes["hazard_rate_active"] * attributes["piT"]
        )

    return attributes


def get_part_count_lambda_b(**kwargs: Dict[str, int]) -> float:
    """Retrieve the parts count base hazard rate (lambda b) from MIL-HDBK-217F.

    This function calculates the MIL-HDBK-217F hazard rate using the parts
    count method.

    This function calculates the MIL-HDBK-217F hazard rate using the parts
    count method.  The dictionary PART_COUNT_217F_LAMBDA_B contains the
    MIL-HDBK-217F parts count base hazard rates.  Keys are for
    PART_COUNT_217F_LAMBDA_B are:

        #. subcategory_id
        #. environment_active_id
        #. type id

    Current subcategory IDs are:

    +----------------+-------------------------------+-----------------+
    | Subcategory    |              Meter            | MIL-HDBK-217F   |
    |       ID       |              Style            |    Section      |
    +================+===============================+=================+
    |        1       | Elapsed Time                  |       12.4      |
    +----------------+-------------------------------+-----------------+
    |        2       | Panel                         |       18.1      |
    +----------------+-------------------------------+-----------------+

    :param subcategory_id: the subcategory identifier.
    :param type_id: the type of meter identifier.
    :param environment_active_id: the active environment identifier.
    :return: _base_hr; the parts count base hazard rate.
    :rtype: float
    :raise: IndexError if passed an unknown active environment ID.
    :raise: KeyError if passed an unknown subcategory ID or type ID.
    """
    _subcategory_id = kwargs.get("subcategory_id", 0)
    _type_id = kwargs.get("type_id", 0)
    _environment_active_id = kwargs.get("environment_active_id", 0)

    return PART_COUNT_LAMBDA_B[_subcategory_id][_type_id][_environment_active_id - 1]


def get_part_stress_lambda_b(subcategory_id: int, type_id: int) -> float:
    """Retrieve the part stress base hazard rate (lambda b) from MIL-HDBK-217F.

    This function calculates the MIL-HDBK-217F hazard rate using the parts
    stress method.

    :param subcategory_id: the subcategory identifier.
    :param type_id: the meter type identifier.
    :return: _lambda_b; the part stress base hazard rate or 0.0 if an unknown
        subcategory ID is passed.
    :rtype: float
    :raise: IndexError when passed an unknown type ID.
    """
    if subcategory_id == 1:
        _lambda_b = PART_STRESS_LAMBDA_B[1][type_id - 1]
    elif subcategory_id == 2:
        _lambda_b = PART_STRESS_LAMBDA_B[2]
    else:
        _lambda_b = 0.0

    return _lambda_b


def get_temperature_stress_factor(
    temperature_active: float, temperature_rated_max: float
) -> float:
    """Retrieve the temperature stress factor (piT).

    :param subcategory_id: the subcategory identifier.
    :param temperature_active: the operating ambient temperature in C.
    :param temperature_rated_max: the maxmimum rated operating
        temperature in C.
    :return: _pi_t; the value of piT associated with the operating temperature.
    :rtype: float
    :raise: TypeError if passed a string for either temperature.
    :raise: ZeroDivisionError if passed a rated maximum temperature = 0.0.
    """
    _temperature_ratio = temperature_active / temperature_rated_max

    if 0.0 < _temperature_ratio <= 0.5:
        _pi_t = 0.5
    elif 0.5 < _temperature_ratio <= 0.6:
        _pi_t = 0.6
    elif 0.6 < _temperature_ratio <= 0.8:
        _pi_t = 0.8
    elif 0.8 < _temperature_ratio <= 1.0:
        _pi_t = 1.0

    return _pi_t



# Standard Library Imports
import logging
from typing import Any, Dict

LOGFORMAT: Any

class RAMSTKLogManager:
    loggers: Dict[str, logging.Logger] = ...
    log_file: Any = ...
    def __init__(self, log_file: str) -> None: ...
    def _do_log_fail_message(self, error_message: str) -> None: ...
    @staticmethod
    def _get_console_handler(log_level: str) -> logging.Handler: ...
    def _get_file_handler(self, log_level: str) -> logging.Handler: ...
    def do_create_logger(
        self, logger_name: str, log_level: str, to_tty: bool = ...
    ) -> None: ...
    def do_log_debug(self, logger_name: str, message: str) -> None: ...
    def do_log_exception(self, logger_name: str, exception: object) -> None: ...
    def do_log_info(self, logger_name: str, message: str) -> None: ...
    def do_log_warning(self, logger_name: str, message: str) -> None: ...
    def do_log_error(self, logger_name: str, message: str) -> None: ...
    def do_log_critical(self, logger_name: str, message: str) -> None: ...



# Standard Library Imports
from typing import Any, Callable, Dict, List, Union

# Third Party Imports
import treelib

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.widgets import RAMSTKComboBox as RAMSTKComboBox
from ramstk.views.gtk3.widgets import RAMSTKFixedPanel as RAMSTKFixedPanel
from ramstk.views.gtk3.widgets import RAMSTKTreePanel as RAMSTKTreePanel

class SimilarItemMethodPanel(RAMSTKFixedPanel):
    _record_field: str
    _select_msg: str
    _tag: str
    _title: str
    cmbSimilarItemMethod: RAMSTKComboBox
    _method_id: int
    _on_edit_message: str
    dic_attribute_widget_map: Dict[str, List[Any]]
    def __init__(self) -> None: ...
    def do_load_comboboxes(self) -> None: ...
    def _on_method_changed(self, combo: RAMSTKComboBox) -> None: ...

class SimilarItemTreePanel(RAMSTKTreePanel):
    _dic_quality: Dict[int, str]
    _dic_environment: Dict[int, str]
    _select_msg: str
    _tag: str
    _title: str
    _dic_hardware_attrs: Dict[int, List[Union[bool, float, int, str]]]
    _dic_reliability_attrs: Dict[int, List[Union[bool, float, int, str]]]
    _dic_row_loader: Dict[str, Callable]
    _dic_visible_mask: Dict[int, Dict[str, str]]
    _method_id: int
    _on_edit_message: str
    dic_attribute_widget_map: Dict[str, List[Any]]
    def __init__(self) -> None: ...
    def do_load_combobox(self) -> None: ...
    def do_refresh_functions(self, row: Gtk.TreeIter, function: List[str]) -> None: ...
    def _do_set_hardware_attributes(self, tree: treelib.Tree) -> None: ...
    def _do_set_reliability_attributes(self, tree: treelib.Tree) -> None: ...
    def _on_method_changed(self, method_id: int) -> None: ...
    _record_id: int
    def _on_row_change(self, selection: Gtk.TreeSelection) -> None: ...
    def _do_set_columns_visible(self) -> None: ...
    def __do_load_similar_item(
        self, node: Any = ..., row: Gtk.TreeIter = ...
    ) -> Gtk.TreeIter: ...



# Standard Library Imports
from typing import Any, Dict, Tuple

# Third Party Imports
import treelib

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gdk as Gdk
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import _ as _
from ramstk.views.gtk3.widgets import RAMSTKButton as RAMSTKButton
from ramstk.views.gtk3.widgets import RAMSTKComboBox as RAMSTKComboBox
from ramstk.views.gtk3.widgets import RAMSTKDateSelect as RAMSTKDateSelect
from ramstk.views.gtk3.widgets import RAMSTKEntry as RAMSTKEntry
from ramstk.views.gtk3.widgets import RAMSTKFixedPanel as RAMSTKFixedPanel
from ramstk.views.gtk3.widgets import RAMSTKSpinButton as RAMSTKSpinButton
from ramstk.views.gtk3.widgets import RAMSTKTextView as RAMSTKTextView
from ramstk.views.gtk3.widgets import RAMSTKTreePanel as RAMSTKTreePanel

class ValidationTreePanel(RAMSTKTreePanel):
    _select_msg: str
    _tag: str
    _title: Any
    _dic_row_loader: Any
    _lst_measurement_units: Any
    _lst_verification_types: Any
    dic_attribute_widget_map: Any
    def __init__(self) -> None: ...
    def do_load_measurement_units(
        self, measurement_unit: Dict[int, Tuple[str, str]]
    ) -> None: ...
    def do_load_verification_types(
        self, verification_type: Dict[int, Tuple[str, str]]
    ) -> None: ...
    def _on_module_switch(self, module: str = ...) -> None: ...
    _record_id: Any
    def _on_row_change(self, selection: Gtk.TreeSelection) -> None: ...
    def __do_load_verification(
        self, node: treelib.Node, row: Gtk.TreeIter
    ) -> Gtk.TreeIter: ...

class ValidationTaskDescriptionPanel(RAMSTKFixedPanel):
    _record_field: str
    _select_msg: str
    _tag: str
    _title: Any
    btnEndDate: Any
    btnStartDate: Any
    cmbTaskType: Any
    cmbMeasurementUnit: Any
    spnStatus: Any
    txtTaskID: Any
    txtCode: Any
    txtMaxAcceptable: Any
    txtMeanAcceptable: Any
    txtMinAcceptable: Any
    txtVarAcceptable: Any
    txtSpecification: Any
    txtTask: Any
    txtEndDate: Any
    txtStartDate: Any
    _dic_task_types: Any
    _dic_units: Any
    dic_attribute_widget_map: Any
    def __init__(self) -> None: ...
    def do_load_measurement_units(
        self, measurement_unit: Dict[int, Tuple[str, str]]
    ) -> None: ...
    def do_load_validation_types(
        self, validation_type: Dict[int, Tuple[str, str]]
    ) -> None: ...
    def _do_make_task_code(self, combo: RAMSTKComboBox) -> None: ...
    @staticmethod
    def _do_select_date(
        __button: RAMSTKButton, __event: Gdk.Event, entry: RAMSTKEntry
    ) -> str: ...

class ValidationTaskEffortPanel(RAMSTKFixedPanel):
    _record_field: str
    _select_msg: str
    _tag: str
    _title: Any
    txtMinTime: Any
    txtExpTime: Any
    txtMaxTime: Any
    txtMinCost: Any
    txtExpCost: Any
    txtMaxCost: Any
    txtMeanTimeLL: Any
    txtMeanTime: Any
    txtMeanTimeUL: Any
    txtMeanCostLL: Any
    txtMeanCost: Any
    txtMeanCostUL: Any
    dic_attribute_widget_map: Any
    def __init__(self) -> None: ...
    def _do_load_code(self, task_code: int) -> None: ...
    def _do_make_task_code(self, task_type: str) -> str: ...
    @staticmethod
    def _do_select_date(
        __button: RAMSTKButton, __event: Gdk.Event, entry: RAMSTKEntry
    ) -> str: ...
    def _on_calculate_task(self, tree: treelib.Tree) -> None: ...
    def __do_adjust_widgets(self) -> None: ...



# Standard Library Imports
from typing import Any, Dict, List, Union

# RAMSTK Package Imports
from ramstk.utilities import none_to_default as none_to_default
from ramstk.views.gtk3 import GObject as GObject
from ramstk.views.gtk3 import Gtk as Gtk

# RAMSTK Local Imports
from .widget import RAMSTKWidget as RAMSTKWidget

class RAMSTKComboBox(Gtk.ComboBox, RAMSTKWidget):
    _default_height: int = ...
    _default_width: int = ...
    _index: Any = ...
    def __init__(self, index: int = ..., simple: bool = ...) -> None: ...
    def do_get_options(self) -> Dict[int, Any]: ...
    def do_load_combo(
        self,
        entries: List[List[Union[str, int]]],
        signal: str = ...,
        simple: bool = ...,
    ) -> None: ...
    def do_update(self, value: int, signal: str = ...) -> None: ...
    def get_value(self, index: int = ...) -> str: ...



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.nswc.record.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTKNSWC Table Module."""

# Third Party Imports
from sqlalchemy import Column, Float, ForeignKey, Integer

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKNSWCRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent ramstk_nswc table in the RAMSTK Program database.

    This table shares a One-to-One relationship with ramstk_hardware.
    """

    __defaults__ = {
        "Cac": 0.0,
        "Calt": 0.0,
        "Cb": 0.0,
        "Cbl": 0.0,
        "Cbt": 0.0,
        "Cbv": 0.0,
        "Cc": 0.0,
        "Ccf": 0.0,
        "Ccp": 0.0,
        "Ccs": 0.0,
        "Ccv": 0.0,
        "Ccw": 0.0,
        "Cd": 0.0,
        "Cdc": 0.0,
        "Cdl": 0.0,
        "Cdp": 0.0,
        "Cds": 0.0,
        "Cdt": 0.0,
        "Cdw": 0.0,
        "Cdy": 0.0,
        "Ce": 0.0,
        "Cf": 0.0,
        "Cg": 0.0,
        "Cga": 0.0,
        "Cgl": 0.0,
        "Cgp": 0.0,
        "Cgs": 0.0,
        "Cgt": 0.0,
        "Cgv": 0.0,
        "Ch": 0.0,
        "Ci": 0.0,
        "Ck": 0.0,
        "Cl": 0.0,
        "Clc": 0.0,
        "Cm": 0.0,
        "Cmu": 0.0,
        "Cn": 0.0,
        "Cnp": 0.0,
        "Cnw": 0.0,
        "Cp": 0.0,
        "Cpd": 0.0,
        "Cpf": 0.0,
        "Cpv": 0.0,
        "Cq": 0.0,
        "Cr": 0.0,
        "Crd": 0.0,
        "Cs": 0.0,
        "Csc": 0.0,
        "Csf": 0.0,
        "Cst": 0.0,
        "Csv": 0.0,
        "Csw": 0.0,
        "Csz": 0.0,
        "Ct": 0.0,
        "Cv": 0.0,
        "Cw": 0.0,
        "Cy": 0.0,
    }
    __tablename__ = "ramstk_nswc"
    __table_args__ = {"extend_existing": True}

    revision_id = Column(
        "fld_revision_id",
        Integer,
        ForeignKey("ramstk_revision.fld_revision_id", ondelete="CASCADE"),
        nullable=False,
    )
    hardware_id = Column(
        "fld_hardware_id",
        Integer,
        ForeignKey("ramstk_hardware.fld_hardware_id", ondelete="CASCADE"),
        primary_key=True,
        nullable=False,
    )

    Cac = Column("fld_c_ac", Float, default=__defaults__["Cac"])
    Calt = Column("fld_c_alt", Float, default=__defaults__["Calt"])
    Cb = Column("fld_c_b", Float, default=__defaults__["Cb"])
    Cbl = Column("fld_c_bl", Float, default=__defaults__["Cbl"])
    Cbt = Column("fld_c_bt", Float, default=__defaults__["Cbt"])
    Cbv = Column("fld_c_bv", Float, default=__defaults__["Cbv"])
    Cc = Column("fld_c_c", Float, default=__defaults__["Cc"])
    Ccf = Column("fld_c_cf", Float, default=__defaults__["Ccf"])
    Ccp = Column("fld_c_cp", Float, default=__defaults__["Ccp"])
    Ccs = Column("fld_c_cs", Float, default=__defaults__["Ccs"])
    Ccv = Column("fld_c_cv", Float, default=__defaults__["Ccv"])
    Ccw = Column("fld_c_cw", Float, default=__defaults__["Ccw"])
    Cd = Column("fld_c_d", Float, default=__defaults__["Cd"])
    Cdc = Column("fld_c_dc", Float, default=__defaults__["Cdc"])
    Cdl = Column("fld_c_dl", Float, default=__defaults__["Cdl"])
    Cdp = Column("fld_c_dp", Float, default=__defaults__["Cdp"])
    Cds = Column("fld_c_ds", Float, default=__defaults__["Cds"])
    Cdt = Column("fld_c_dt", Float, default=__defaults__["Cdt"])
    Cdw = Column("fld_c_dw", Float, default=__defaults__["Cdw"])
    Cdy = Column("fld_c_dy", Float, default=__defaults__["Cdy"])
    Ce = Column("fld_c_e", Float, default=__defaults__["Ce"])
    Cf = Column("fld_c_f", Float, default=__defaults__["Cf"])
    Cg = Column("fld_c_g", Float, default=__defaults__["Cg"])
    Cga = Column("fld_c_ga", Float, default=__defaults__["Cga"])
    Cgl = Column("fld_c_gl", Float, default=__defaults__["Cgl"])
    Cgp = Column("fld_c_gp", Float, default=__defaults__["Cgp"])
    Cgs = Column("fld_c_gs", Float, default=__defaults__["Cgs"])
    Cgt = Column("fld_c_gt", Float, default=__defaults__["Cgt"])
    Cgv = Column("fld_c_gv", Float, default=__defaults__["Cgv"])
    Ch = Column("fld_c_h", Float, default=__defaults__["Ch"])
    Ci = Column("fld_c_i", Float, default=__defaults__["Ci"])
    Ck = Column("fld_c_k", Float, default=__defaults__["Ck"])
    Cl = Column("fld_c_l", Float, default=__defaults__["Cl"])
    Clc = Column("fld_c_lc", Float, default=__defaults__["Clc"])
    Cm = Column("fld_c_m", Float, default=__defaults__["Cm"])
    Cmu = Column("fld_c_mu", Float, default=__defaults__["Cmu"])
    Cn = Column("fld_c_n", Float, default=__defaults__["Cn"])
    Cnp = Column("fld_c_np", Float, default=__defaults__["Cnp"])
    Cnw = Column("fld_c_nw", Float, default=__defaults__["Cnw"])
    Cp = Column("fld_c_p", Float, default=__defaults__["Cp"])
    Cpd = Column("fld_c_pd", Float, default=__defaults__["Cpd"])
    Cpf = Column("fld_c_pf", Float, default=__defaults__["Cpf"])
    Cpv = Column("fld_c_pv", Float, default=__defaults__["Cpv"])
    Cq = Column("fld_c_q", Float, default=__defaults__["Cq"])
    Cr = Column("fld_c_r", Float, default=__defaults__["Cr"])
    Crd = Column("fld_c_rd", Float, default=__defaults__["Crd"])
    Cs = Column("fld_c_s", Float, default=__defaults__["Cs"])
    Csc = Column("fld_c_sc", Float, default=__defaults__["Csc"])
    Csf = Column("fld_c_sf", Float, default=__defaults__["Csf"])
    Cst = Column("fld_c_st", Float, default=__defaults__["Cst"])
    Csv = Column("fld_c_sv", Float, default=__defaults__["Csv"])
    Csw = Column("fld_c_sw", Float, default=__defaults__["Csw"])
    Csz = Column("fld_c_sz", Float, default=__defaults__["Csz"])
    Ct = Column("fld_c_t", Float, default=__defaults__["Ct"])
    Cv = Column("fld_c_v", Float, default=__defaults__["Cv"])
    Cw = Column("fld_c_w", Float, default=__defaults__["Cw"])
    Cy = Column("fld_c_y", Float, default=__defaults__["Cy"])

    # Define the relationships to other tables in the RAMSTK Program database.

    def get_attributes(self):
        """Retrieve the current values of the RAMSTKNSWC data model attributes.

        :return: {hardware_id, Cac, Calt, Cb, Cbl, Cbt, Cbv, Cc, Ccf, Ccp, Ccs,
                  Ccv, Ccw, Cd, Cdc, Cdl, Cdp, Cds, Cdw, Cdy, Ce, Cf, Cg, Cga,
                  Cgl, Cgp, Cgs, Cgt, Cgv, Ch, Ci, Ck, Cl, Clc, Cm, Cmu, Cn,
                  Cnp, Cnw, Cp, Cpd, Cpf, Cpv, Cq, Cr, Crd, Cs, Csc, Csf, Cst,
                  Csv, Csw, Csz, Ct, Cv, Cw, Cy} pairs.
        :rtype: dict
        """
        _attributes = {
            "hardware_id": self.hardware_id,
            "Cac": self.Cac,
            "Calt": self.Calt,
            "Cb": self.Cb,
            "Cbl": self.Cbl,
            "Cbt": self.Cbt,
            "Cbv": self.Cbv,
            "Cc": self.Cc,
            "Ccf": self.Ccf,
            "Ccp": self.Ccp,
            "Ccs": self.Ccs,
            "Ccv": self.Ccv,
            "Ccw": self.Ccw,
            "Cd": self.Cd,
            "Cdc": self.Cdc,
            "Cdl": self.Cdl,
            "Cdp": self.Cdp,
            "Cds": self.Cds,
            "Cdt": self.Cdt,
            "Cdw": self.Cdw,
            "Cdy": self.Cdy,
            "Ce": self.Ce,
            "Cf": self.Cf,
            "Cg": self.Cg,
            "Cga": self.Cga,
            "Cgl": self.Cgl,
            "Cgp": self.Cgp,
            "Cgs": self.Cgs,
            "Cgt": self.Cgt,
            "Cgv": self.Cgv,
            "Ch": self.Ch,
            "Ci": self.Ci,
            "Ck": self.Ck,
            "Cl": self.Cl,
            "Clc": self.Clc,
            "Cm": self.Cm,
            "Cmu": self.Cmu,
            "Cn": self.Cn,
            "Cnp": self.Cnp,
            "Cnw": self.Cnw,
            "Cp": self.Cp,
            "Cpd": self.Cpd,
            "Cpf": self.Cpf,
            "Cpv": self.Cpv,
            "Cq": self.Cq,
            "Cr": self.Cr,
            "Crd": self.Crd,
            "Cs": self.Cs,
            "Csc": self.Csc,
            "Csf": self.Csf,
            "Cst": self.Cst,
            "Csv": self.Csv,
            "Csw": self.Csw,
            "Csz": self.Csz,
            "Ct": self.Ct,
            "Cv": self.Cv,
            "Cw": self.Cw,
            "Cy": self.Cy,
        }

        return _attributes



# Standard Library Imports
from typing import Any

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE as RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord as RAMSTKBaseRecord

class RAMSTKDesignMechanicRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    __defaults__: Any
    __tablename__: str
    __table_args__: Any
    revision_id: Any
    hardware_id: Any
    altitude_operating: Any
    application_id: Any
    balance_id: Any
    clearance: Any
    casing_id: Any
    contact_pressure: Any
    deflection: Any
    diameter_coil: Any
    diameter_inner: Any
    diameter_outer: Any
    diameter_wire: Any
    filter_size: Any
    flow_design: Any
    flow_operating: Any
    frequency_operating: Any
    friction: Any
    impact_id: Any
    leakage_allowable: Any
    length: Any
    length_compressed: Any
    length_relaxed: Any
    load_design: Any
    load_id: Any
    load_operating: Any
    lubrication_id: Any
    manufacturing_id: Any
    material_id: Any
    meyer_hardness: Any
    misalignment_angle: Any
    n_ten: Any
    n_cycles: Any
    n_elements: Any
    offset: Any
    particle_size: Any
    pressure_contact: Any
    pressure_delta: Any
    pressure_downstream: Any
    pressure_rated: Any
    pressure_upstream: Any
    rpm_design: Any
    rpm_operating: Any
    service_id: Any
    spring_index: Any
    surface_finish: Any
    technology_id: Any
    thickness: Any
    torque_id: Any
    type_id: Any
    viscosity_design: Any
    viscosity_dynamic: Any
    water_per_cent: Any
    width_minimum: Any
    def get_attributes(self): ...



# Standard Library Imports
from typing import Any, Dict

def _calculate_agree_apportionment(
    parent_goal: float, attributes: Dict[str, Any]
) -> Dict[str, Any]: ...
def _calculate_arinc_apportionment(
    parent_goal: float, attributes: Dict[str, Any]
) -> Dict[str, Any]: ...
def _calculate_equal_apportionment(
    parent_goal: float, attributes: Dict[str, Any]
) -> Dict[str, Any]: ...
def _calculate_foo_apportionment(
    parent_goal: float, cum_weight: int, attributes: Dict[str, Any]
) -> Dict[str, Any]: ...
def _from_hazard_rate_goal(attributes: Dict[str, Any]) -> Dict[str, Any]: ...
def _from_mtbf_goal(attributes: Dict[str, Any]) -> Dict[str, Any]: ...
def _from_reliability_goal(attributes: Dict[str, Any]) -> Dict[str, Any]: ...
def do_allocate_reliability(
    parent_goal: float, cumulative_weight: int, **attributes: Dict[str, Any]
) -> Dict[str, Any]: ...
def do_calculate_goals(**attributes: Dict[str, Any]) -> Dict[str, Any]: ...
def get_allocation_goal(**attributes: Dict[str, Any]) -> Dict[str, Any]: ...



# -*- coding: utf-8 -*-
#
#       ramstk.analyses.prediction.Fuse.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Fuse MIL-HDBK-217F Constants and Calculations Module."""

# Standard Library Imports
from typing import Any, Dict

PART_COUNT_LAMBDA_B = [
    0.01,
    0.02,
    0.06,
    0.05,
    0.11,
    0.09,
    0.12,
    0.15,
    0.18,
    0.18,
    0.009,
    0.1,
    0.21,
    2.3,
]
PI_E = [1.0, 2.0, 8.0, 5.0, 11.0, 9.0, 12.0, 15.0, 18.0, 16.0, 0.9, 10.0, 21.0, 230.0]


def calculate_part_count(**attributes: Dict[str, Any]) -> float:
    """Wrap get_part_count_lambda_b().

    This wrapper allows us to pass an attributes dict from a generic parts
    count function.

    :param attributes: the attributes for the fuse being calculated.
    :return: _base_hr; the parts count base hazard rates.
    :rtype: float
    """
    return get_part_count_lambda_b(attributes["environment_active_id"])  # type: ignore


def calculate_part_stress(**attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the part stress hazard rate for a fuse.

    This function calculates the MIL-HDBK-217F hazard rate using the part
    stress method.

    :return: attributes; the keyword argument (hardware attribute) dictionary
        with updated values.
    :rtype: dict
    """
    attributes["hazard_rate_active"] = 0.010 * attributes["piE"]  # type: ignore

    return attributes


def get_part_count_lambda_b(environment_active_id: int) -> float:
    """Retrieve the part count hazard rate for a fuse.

    :param environment_active_id: the active environment identifer.
    :return: _base_hr; the part count base hazard rate.
    :rtype: float
    :raise: IndexError when passed an unkown active environment ID.
    """
    return PART_COUNT_LAMBDA_B[environment_active_id - 1]



# pylint: disable=duplicate-code
# -*- coding: utf-8 -*-
#
#       ramstk.models.programdb.RAMSTKMission.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Mission Record Module."""

# Third Party Imports
from sqlalchemy import Column, Float, ForeignKey, Integer, String, UniqueConstraint
from sqlalchemy.orm import relationship

# RAMSTK Package Imports
from ramstk.db import RAMSTK_BASE
from ramstk.models import RAMSTKBaseRecord


class RAMSTKMissionRecord(RAMSTK_BASE, RAMSTKBaseRecord):
    """Class to represent ramstk_mission table in the RAMSTK Program database.

    This table shares a Many-to-One relationship with ramstk_revision. This table
    shares a One-to-Many relationship with ramstk_mission_phase.
    """

    __defaults__ = {"description": "", "mission_time": 0.0, "time_units": "hours"}
    __tablename__ = "ramstk_mission"
    __table_args__ = (
        UniqueConstraint(
            "fld_revision_id", "fld_mission_id", name="ramstk_mission_ukey"
        ),
        {"extend_existing": True},
    )

    revision_id = Column(
        "fld_revision_id",
        Integer,
        ForeignKey("ramstk_revision.fld_revision_id", ondelete="CASCADE"),
        nullable=False,
    )
    mission_id = Column(
        "fld_mission_id",
        Integer,
        primary_key=True,
        autoincrement=True,
        nullable=False,
    )
    description = Column("fld_description", String, default=__defaults__["description"])
    mission_time = Column(
        "fld_mission_time", Float, default=__defaults__["mission_time"]
    )
    time_units = Column(
        "fld_time_units", String(256), default=__defaults__["time_units"]
    )

    # Define the relationships to other tables in the RAMSTK Program database.
    phase = relationship(  # type: ignore
        "RAMSTKMissionPhaseRecord",
        back_populates="mission",
        cascade="all,delete",
    )

    is_mission = True
    is_phase = False
    is_env = False

    def get_attributes(self):
        """Retrieve current values of the RAMSTKMission data model attributes.

        :return: (revision_id, mission_id, description, mission_time,
                  time_units)
        :rtype: tuple
        """
        _attributes = {
            "revision_id": self.revision_id,
            "mission_id": self.mission_id,
            "description": self.description,
            "mission_time": self.mission_time,
            "time_units": self.time_units,
        }

        return _attributes



# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.books.workbook.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright since 2007 Doyle "weibullguy" Rowland doyle.rowland <AT> reliaqual <DOT> com
"""GTK3 Work Book."""

# Standard Library Imports
from typing import Dict, List

# Third Party Imports
from pubsub import pub

# RAMSTK Package Imports
from ramstk.configuration import RAMSTKUserConfiguration
from ramstk.logger import RAMSTKLogManager
from ramstk.views.gtk3.allocation import AllocationWorkView
from ramstk.views.gtk3.fmea import FMEAWorkView
from ramstk.views.gtk3.function import FunctionWorkView
from ramstk.views.gtk3.hardware import (
    HardwareAssessmentInputView,
    HardwareAssessmentResultsView,
    HardwareGeneralDataView,
)
from ramstk.views.gtk3.hazard_analysis import HazardsWorkView
from ramstk.views.gtk3.pof import PoFWorkView
from ramstk.views.gtk3.program_status import ProgramStatusWorkView
from ramstk.views.gtk3.requirement import (
    RequirementAnalysisView,
    RequirementGeneralDataView,
)
from ramstk.views.gtk3.revision import RevisionWorkView
from ramstk.views.gtk3.similar_item import SimilarItemWorkView
from ramstk.views.gtk3.validation import ValidationGeneralDataView
from ramstk.views.gtk3.widgets import RAMSTKBaseBook, RAMSTKBaseView


class RAMSTKWorkBook(RAMSTKBaseBook):
    """The Work Book for the pyGObject (GTK3) interface."""

    def __init__(
        self, configuration: RAMSTKUserConfiguration, logger: RAMSTKLogManager
    ) -> None:
        """Initialize an instance of the Work View class.

        :param configuration: the RAMSTKUserConfiguration class instance.
        :type configuration:
            :class:`ramstk.configuration.RAMSTKUserConfiguration`
        :param logger: the RAMSTKLogManager class instance.
        :type logger: :class:`ramstk.logger.RAMSTKLogManager`
        """
        super().__init__(configuration)

        # Initialize private dictionary attributes.

        # Initialize private list attributes.

        # Initialize private scalar attributes.

        # Initialize public dictionary attributes.
        self.dic_work_views: Dict[str, List[RAMSTKBaseView]] = {
            "revision": [RevisionWorkView(configuration, logger)],
            "function": [
                FunctionWorkView(configuration, logger),
                HazardsWorkView(configuration, logger),
            ],
            "requirement": [
                RequirementGeneralDataView(configuration, logger),
                RequirementAnalysisView(configuration, logger),
            ],
            "hardware": [
                HardwareGeneralDataView(configuration, logger),
                AllocationWorkView(configuration, logger),
                SimilarItemWorkView(configuration, logger),
                HardwareAssessmentInputView(configuration, logger),
                HardwareAssessmentResultsView(configuration, logger),
                FMEAWorkView(configuration, logger),
                PoFWorkView(configuration, logger),
            ],
            "validation": [
                ValidationGeneralDataView(configuration, logger),
                ProgramStatusWorkView(configuration, logger),
            ],
        }

        # Initialize public list attributes.

        # Initialize public scalar attributes.

        self._set_properties("workbook")

        # Subscribe to PyPubSub messages.
        pub.subscribe(self._on_module_change, "mvwSwitchedPage")

    def _on_module_change(self, module: str = "") -> None:
        """Load Work Views for the RAMSTK module selected in the Module Book.

        :return: None
        :rtype: None
        """
        for _page in self.get_children():
            self.remove(_page)

        try:
            for _workspace in self.dic_work_views[module]:
                self.insert_page(_workspace, _workspace.hbx_tab_label, -1)
        except KeyError:
            pass



# Standard Library Imports
from typing import Any

# RAMSTK Package Imports
from ramstk.views.gtk3 import GObject as GObject
from ramstk.views.gtk3 import _ as _

class RAMSTKWidget:
    _default_height: int = ...
    _default_width: int = ...
    dic_handler_id: Any = ...
    height: int = ...
    width: int = ...
    def __init__(self) -> None: ...
    def do_set_properties(self, **kwargs: Any) -> None: ...



# Standard Library Imports
from typing import Any, Callable

# RAMSTK Package Imports
from ramstk.views.gtk3 import Gtk as Gtk
from ramstk.views.gtk3 import Pango as Pango

# RAMSTK Local Imports
from .widget import RAMSTKWidget as RAMSTKWidget

class RAMSTKEntry(Gtk.Entry, RAMSTKWidget):
    _default_height: int
    _default_width: int
    def __init__(self) -> None: ...
    def do_get_text(self): ...
    def do_set_properties(self, **kwargs: Any) -> None: ...
    def do_update(self, value: Any, signal: str = ...) -> None: ...

class RAMSTKTextView(Gtk.TextView, RAMSTKWidget):
    _default_height: int
    _default_width: int
    scrollwindow: Any
    tag_bold: Any
    def __init__(self, txvbuffer: Gtk.TextBuffer) -> None: ...
    def connect(
        self, signal: str, callback: Callable, index: int, message: str
    ) -> int: ...
    def do_get_buffer(self) -> Gtk.TextBuffer: ...
    def do_get_text(self) -> Any: ...
    def do_set_properties(self, **kwargs: Any) -> None: ...
    def do_update(self, value: str, signal: str = ...) -> None: ...



# -*- coding: utf-8 -*-
#
#       ramstk.analyses.prediction.Switch.py is part of the RAMSTK Project
#
# All rights reserved.
# Copyright 2007 - 2017 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""Switch Reliability Calculations Module."""

# Standard Library Imports
from math import exp
from typing import Any, Dict, List

PART_COUNT_LAMBDA_B: Dict[int, List[float]] = {
    1: [
        0.0010,
        0.0030,
        0.018,
        0.0080,
        0.029,
        0.010,
        0.018,
        0.013,
        0.022,
        0.046,
        0.0005,
        0.025,
        0.067,
        1.2,
    ],
    2: [0.15, 0.44, 2.7, 1.2, 4.3, 1.5, 2.7, 1.9, 3.3, 6.8, 0.74, 3.7, 9.9, 180.0],
    3: [0.33, 0.99, 5.9, 2.6, 9.5, 3.3, 5.9, 4.3, 7.2, 15.0, 0.16, 8.2, 22.0, 390.0],
    4: [
        0.56,
        1.7,
        10.0,
        4.5,
        16.0,
        5.6,
        10.0,
        7.3,
        12.0,
        26.0,
        0.26,
        14.0,
        38.0,
        670.0,
    ],
}
PART_COUNT_LAMBDA_B_BREAKER: Dict[int, List[float]] = {
    1: [0.11, 0.23, 1.7, 0.91, 3.1, 0.8, 1.0, 1.3, 1.4, 5.2, 0.057, 2.8, 7.5, 0.0],
    2: [
        0.060,
        0.12,
        0.90,
        0.48,
        1.6,
        0.42,
        0.54,
        0.66,
        0.72,
        2.8,
        0.030,
        1.5,
        4.0,
        0.0,
    ],
}

PART_COUNT_PI_Q = {
    1: [1.0, 20.0],
    2: [1.0, 20.0],
    3: [1.0, 50.0],
    4: [1.0, 10.0],
    5: [1.0, 8.4],
}
PART_STRESS_LAMBDA_B_TOGGLE: Dict[int, List[float]] = {
    1: [0.00045, 0.034],
    2: [0.0027, 0.04],
}
PART_STRESS_LAMBDA_B_BREAKER: List[float] = [0.02, 0.038, 0.038]
PART_STRESS_PI_Q = {5: [1.0, 8.4]}
PI_C = {1: [1.0, 1.5, 1.7, 2.0, 2.5, 3.0, 4.2, 5.5, 8.0], 5: [1.0, 2.0, 3.0, 4.0]}
PI_E = {
    1: [
        1.0,
        3.0,
        18.0,
        8.0,
        29.0,
        10.0,
        18.0,
        13.0,
        22.0,
        46.0,
        0.5,
        25.0,
        67.0,
        1200.0,
    ],
    2: [
        1.0,
        3.0,
        18.0,
        8.0,
        29.0,
        10.0,
        18.0,
        13.0,
        22.0,
        46.0,
        0.5,
        25.0,
        67.0,
        1200.0,
    ],
    3: [
        1.0,
        3.0,
        18.0,
        8.0,
        29.0,
        10.0,
        18.0,
        13.0,
        22.0,
        46.0,
        0.5,
        25.0,
        67.0,
        1200.0,
    ],
    4: [
        1.0,
        3.0,
        18.0,
        8.0,
        29.0,
        10.0,
        18.0,
        13.0,
        22.0,
        46.0,
        0.5,
        25.0,
        67.0,
        1200.0,
    ],
    5: [1.0, 2.0, 15.0, 8.0, 27.0, 7.0, 9.0, 11.0, 12.0, 46.0, 0.5, 25.0, 67.0, 0.0],
}


def calculate_load_stress_factor(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the load stress factor (piL).

    :param attributes: the attributes of the switch being calculated.
    :return attributes: the updated attributes of the switch being calculated.
    :rtype: dict
    """
    application_id: Any = attributes["application_id"]
    current_ratio: Any = attributes["current_ratio"]
    _pi_l: Any = 0.0

    if application_id == 1:  # Resistive
        _pi_l = exp((current_ratio / 0.8) ** 2.0)
    elif application_id == 2:  # Inductive
        _pi_l = exp((current_ratio / 0.4) ** 2.0)
    elif application_id == 3:  # Capacitive
        _pi_l = exp((current_ratio / 0.2) ** 2.0)

    attributes["piL"] = _pi_l

    return attributes


def calculate_part_count(**attributes: Dict[str, Any]) -> float:
    """Wrap get_part_count_lambda_b().

    This wrapper allows us to pass an attributes dict from a generic parts
    count function.

    :param attributes: the attributes for the connection being calculated.
    :return: _base_hr; the parts count base hazard rates.
    :rtype: float
    """
    return get_part_count_lambda_b(attributes)


def calculate_part_stress_lambda_b(attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate part stress base hazard rate (lambda b) from MIL-HDBK-217F.

    This function calculates the MIL-HDBK-217F hazard rate using the parts
    stress method.

    :param attributes: the attributes of the switch being calculated.
    :return attributes: the updated attributes of the switch being calculated.
    :rtype: dict
    :raise: IndexError if passed an unknown quality ID or application ID.
    :raise: KeyError is passed an unknown construction ID.
    """
    _subcategory_id: Any = attributes["subcategory_id"]
    _quality_id: Any = attributes["quality_id"]
    _construction_id: Any = attributes["construction_id"]
    _application_id: Any = attributes["application_id"]
    _n_elements: Any = attributes["n_elements"]

    _dic_factors: Dict[int, List[List[float]]] = {
        2: [[0.1, 0.00045, 0.0009], [0.1, 0.23, 0.63]],
        3: [[0.0067, 0.00003, 0.00003], [0.1, 0.02, 0.06]],
        4: [[0.0067, 0.062], [0.086, 0.089]],
    }

    if _subcategory_id == 1:
        _lambda_b: Any = PART_STRESS_LAMBDA_B_TOGGLE[_construction_id][_quality_id - 1]
    elif _subcategory_id in [2, 3]:
        _lambda_bE = _dic_factors[_subcategory_id][_quality_id - 1][0]
        _lambda_bC = _dic_factors[_subcategory_id][_quality_id - 1][1]
        _lambda_b0 = _dic_factors[_subcategory_id][_quality_id - 1][2]
        if _construction_id == 1:
            _lambda_b = _lambda_bE + _n_elements * _lambda_bC
        else:
            _lambda_b = _lambda_bE + _n_elements * _lambda_b0
    elif _subcategory_id == 4:
        _lambda_b1 = _dic_factors[_subcategory_id][_quality_id - 1][0]
        _lambda_b2 = _dic_factors[_subcategory_id][_quality_id - 1][1]
        _lambda_b = _lambda_b1 + _n_elements * _lambda_b2
    elif _subcategory_id == 5:
        _lambda_b = PART_STRESS_LAMBDA_B_BREAKER[_application_id - 1]
    else:
        _lambda_b = 0.0

    attributes["lambda_b"] = _lambda_b

    return attributes


def calculate_part_stress(**attributes: Dict[str, Any]) -> Dict[str, Any]:
    """Calculate the part stress hazard rate for a switch.

    This function calculates the MIL-HDBK-217F hazard rate using the part
    stress method.

    :param attributes: the attributes of the switch being calculated.
    :return attributes: the updated attributes of the switch being calculated.
    :rtype: dict
    """
    _subcategory_id: Any = attributes["subcategory_id"]
    _application_id: Any = attributes["application_id"]
    _n_cycles: Any = attributes["n_cycles"]
    _pi_e: Any = attributes["piE"]
    _pi_q: Any = attributes["piQ"]
    _pi_c: Any = 1.0
    _pi_cyc: Any = 1.0

    attributes = calculate_part_stress_lambda_b(attributes)
    attributes = calculate_load_stress_factor(attributes)
    _lambda_b: Any = attributes["lambda_b"]
    _pi_l: Any = attributes["piL"]

    # Determine the contact form and quantity factor (piC).
    if _subcategory_id in [1, 5]:
        _contact_form_id: Any = attributes["contact_form_id"]
        _pi_c = PI_C[_subcategory_id][_contact_form_id]
        attributes["piC"] = _pi_c

    # Determine the cycling factor (piCYC).
    if _n_cycles > 1:
        _pi_cyc = float(_n_cycles)
    attributes["piCYC"] = _pi_cyc

    # Determine the use factor (piU).
    _pi_u: Any = 10.0 if _application_id - 1 else 1.0
    attributes["piU"] = _pi_u

    _hazard_rate_active = _lambda_b * _pi_e
    if _subcategory_id == 1:
        _hazard_rate_active = _hazard_rate_active * _pi_cyc * _pi_l * _pi_c
    elif _subcategory_id in [2, 3, 4]:
        _hazard_rate_active = _hazard_rate_active * _pi_cyc * _pi_l
    elif _subcategory_id == 5:
        _hazard_rate_active = _hazard_rate_active * _pi_c * _pi_u * _pi_q

    attributes["hazard_rate_active"] = _hazard_rate_active

    return attributes


def get_part_count_lambda_b(attributes: Dict[str, Any]) -> float:
    """Retrieve parts count base hazard rate (lambda b) from MIL-HDBK-217F.

    This function calculates the MIL-HDBK-217F hazard rate using the parts
    count method.

    This function calculates the MIL-HDBK-217F hazard rate using the parts
    count method.  The dictionary PART_COUNT_217F_LAMBDA_B contains the
    MIL-HDBK-217F parts count base hazard rates.  Keys are for
    PART_COUNT_217F_LAMBDA_B are:

        #. subcategory_id
        #. environment_active_id
        #. construction_id; if the switch subcategory is NOT construction
            dependent, then the second key will be zero.

    Current subcategory IDs are:

    +----------------+-------------------------------+-----------------+
    | Subcategory    |             Switch            | MIL-HDBK-217F   |
    |       ID       |              Style            |    Section      |
    +================+===============================+=================+
    |        1       | Toggle or Pushbutton          |       15.1      |
    +----------------+-------------------------------+-----------------+
    |        2       | Sensitive                     |       15.2      |
    +----------------+-------------------------------+-----------------+
    |        3       | Rotary                        |       15.3      |
    +----------------+-------------------------------+-----------------+
    |        4       | Thumbwheel                    |       16.1      |
    +----------------+-------------------------------+-----------------+
    |        5       | Circuit Breaker               |       17.1      |
    +----------------+-------------------------------+-----------------+

    :param attributes: the attributes of the switch being calculated.
    :return: _base_hr; the parts count base hazard rate.
    :rtype: float
    :raise: IndexError if passed an unknown active environment ID.
    :raise: KeyError if passed an unknown subcategory ID or construction ID.
    """
    _construction_id: Any = attributes["construction_id"]
    _environment_active_id: Any = attributes["environment_active_id"]
    _subcategory_id: Any = attributes["subcategory_id"]

    if _subcategory_id == 5:
        _base_hr: float = PART_COUNT_LAMBDA_B_BREAKER[_construction_id][
            _environment_active_id - 1
        ]
    else:
        _base_hr = PART_COUNT_LAMBDA_B[_subcategory_id][_environment_active_id - 1]

    return _base_hr



# pylint: disable=non-parent-init-called
# -*- coding: utf-8 -*-
#
#       ramstk.views.gtk3.widgets.scrolledwindow.py is part of the RAMSTK
#       Project
#
# All rights reserved.
# Copyright 2007 - 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK GTK3 ScrolledWindow Module."""

# RAMSTK Package Imports
from ramstk.views.gtk3 import GObject, Gtk


class RAMSTKScrolledWindow(Gtk.ScrolledWindow):
    """The RAMSTKScrolledWindow class."""

    def __init__(self, child: object) -> None:
        """Create ScrolledWindow() widgets.

        :param child: the Gtk.Widget() to add to the scrolled window.
        """
        GObject.GObject.__init__(self)

        self.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)

        if child is not None:
            self.add(child)



# -*- coding: utf-8 -*-
#
#       ramstk.db.base.py is part of The RAMSTK Project
#
# All rights reserved.
# Copyright 2019 Doyle Rowland doyle.rowland <AT> reliaqual <DOT> com
"""RAMSTK Base Database Module."""

# Standard Library Imports
import sqlite3
from typing import Any, Dict, List, TextIO, Tuple

# Third Party Imports
import psycopg2  # type: ignore
from psycopg2 import sql  # type: ignore
from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT  # type: ignore
from pubsub import pub

# noinspection PyPackageRequirements
from sqlalchemy import create_engine, exc

# noinspection PyPackageRequirements,PyProtectedMember
from sqlalchemy.engine import Engine  # type: ignore

# noinspection PyPackageRequirements
from sqlalchemy.orm import query, scoped_session, sessionmaker  # type: ignore
from sqlalchemy.orm.exc import FlushError  # type: ignore

# RAMSTK Package Imports
from ramstk.exceptions import DataAccessError


def do_create_program_db(database: Dict[str, str], sql_file: TextIO) -> None:
    """Create a shiny new, unpopulated RAMSTK program database.

    :param database: a dict containing the database connection arguments.
    :param sql_file: the absolute path to the text file containing the
        SQL statements for creating a bare RAMSTK Program Database.
    :return: None
    :rtype: None
    """
    conn: Any = ""

    if database["dialect"] == "sqlite":
        conn = sqlite3.connect(database["database"])
        conn.executescript(sql_file.read().strip())
    elif database["dialect"] == "postgres":
        # Create the database.
        conn = psycopg2.connect(
            host=database["host"],
            dbname="postgres",
            user=database["user"],
            # deepcode ignore NoHardcodedPasswords:
            password=database["password"],
        )
        conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)

        cursor = conn.cursor()
        cursor.execute(
            sql.SQL("DROP DATABASE IF EXISTS {}").format(
                sql.Identifier(database["database"])
            )
        )
        cursor.execute(
            sql.SQL("CREATE DATABASE {}").format(sql.Identifier(database["database"]))
        )
        cursor.close()
        conn.close()

        # Populate the database.
        conn = psycopg2.connect(
            host=database["host"],
            dbname=database["database"],
            user=database["user"],
            # deepcode ignore NoHardcodedPasswords:
            password=database["password"],
        )
        conn.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        conn.set_session(autocommit=True)

        cursor = conn.cursor()
        cursor.execute(sql_file.read())
        cursor.close()

    conn.close()


def do_open_session(database: str) -> Tuple[Engine, scoped_session]:
    """Create a session to be used with an instance of the BaseDatabase."""
    engine: Any = create_engine(database)
    # deepcode ignore missing~close~connect: engines are disposed
    engine.connect()

    return (
        engine,
        scoped_session(sessionmaker(autocommit=False, autoflush=False, bind=engine)),
    )


# noinspection PyUnresolvedReferences
class BaseDatabase:
    """The BaseDatabase class."""

    # Define public class dict attributes.
    cxnargs: Dict[str, str] = {
        "dialect": "",
        "user": "",
        "password": "",
        "host": "",
        "port": "",
        "dbname": "",
    }

    # Define public class scalar attributes.
    engine: Engine = None  # type: ignore
    session: scoped_session = None  # type: ignore
    database: str = ""
    sqlstatements: Dict[str, str] = {
        "select": "SELECT {0:s} ",
        "from": "FROM {0:s} ",
        "order": "ORDER BY {0:s} DESC LIMIT 1",
    }

    def __init__(self) -> None:
        """Initialize an instance of the BaseDatabase."""

        # Initialize private dictionary instance attributes.

        # Initialize private list instance attributes.

        # Initialize private scalar instance attributes.

        # Initialize public dictionary instance attributes.

        # Initialize public list instance attributes.

        # Initialize public scalar instance attributes.

    def do_connect(self, database: Dict) -> None:
        """Connect to the database.

        :param database: the connection information for the database to
            connect to.
        :return: None
        :rtype: None
        :raise: sqlalchemy.exc.OperationalError if passed an invalid database
            URL.
        :raise: sqlalchemy.exc.ArgumentError if passed a database URL with
            an unknown/unsupported SQL dialect.
        """
        self.cxnargs["dialect"] = database["dialect"]
        self.cxnargs["user"] = database["user"]
        self.cxnargs["password"] = database["password"]
        self.cxnargs["host"] = database["host"]
        self.cxnargs["port"] = database["port"]
        self.cxnargs["dbname"] = database["database"]

        try:
            if self.cxnargs["dialect"] == "sqlite":
                self.database = "sqlite:///" + self.cxnargs["dbname"]
            elif self.cxnargs["dialect"] == "postgres":
                self.database = (
                    "postgresql+psycopg2://"
                    + self.cxnargs["user"]
                    + ":"
                    + self.cxnargs["password"]
                    + "@"
                    + self.cxnargs["host"]
                    + ":"
                    + self.cxnargs["port"]
                    + "/"
                    + self.cxnargs["dbname"]
                )
            else:
                raise DataAccessError(
                    "Unknown database dialect in database " "connection dict."
                )
        except TypeError as _error:
            raise DataAccessError(
                "Unknown dialect or non-string value in " "database connection dict."
            ) from _error

        if self.database != "":
            self.engine, self.session = do_open_session(self.database)

    def do_delete(self, item: object) -> None:
        """Delete a record from the RAMSTK Program database.

        :param item: the item to remove from the RAMSTK Program database.
        :type item: Object()
        :return: None
        :rtype: None
        """
        try:
            self.session.delete(item)
            self.session.commit()
        except exc.InvalidRequestError as _error:
            # This exception generally corresponds to runtime state errors.
            # These types of errors are unlikely to be user errors and will
            # most likely be the result of a corrupted database.  Some
            # situations that can raise this exception are:
            #   1. Attempting to delete a record from a non-existent table.
            self.session.rollback()
            _error_message = (
                "There was an database error when attempting to delete a "
                "record.  Error returned from database was:\n\t{0:s}.".format(
                    str(_error)
                )
            )
            pub.sendMessage("fail_delete_record", error_message=_error_message)
            raise DataAccessError(_error_message) from _error
        except exc.ProgrammingError as _error:
            # This exception is raised when there is an error during
            # execution of a SQL statement.  These types of errors are
            # unlikely to be user errors and will most likely be the result of
            # a corrupted database.  Some situations that can raise this
            # exception are:
            #   1. Foreign key exists, but foreign table does not.
            self.session.rollback()
            _error_message = (
                "There was an database error when attempting to delete a "
                "record.  Error returned from database was:\n\t{0:s}.".format(
                    str(_error.orig)
                )
            )
            pub.sendMessage("fail_delete_record", error_message=_error_message)
            raise DataAccessError(_error_message) from _error

    def do_disconnect(self) -> None:
        """Close the current session.

        :return: None
        :rtype: None
        """
        self.session.close()
        self.engine.dispose()
        # noinspection PyTypeChecker
        self.session = None  # type: ignore
        self.database = ""

    def do_insert(self, record: object) -> None:
        """Add a new record to a database table.

        :param record: the object to add to the RAMSTK Program database.
        :return: None
        :rtype: None
        """
        try:
            self.session.add(record)
            self.session.commit()
        except AttributeError as _error:
            # This exception is raised when there is no database connection.
            _error_message = (
                "dao.do_insert: No database connected when attempting to add a record."
            )
            pub.sendMessage(
                "fail_insert_record",
                error_message=_error_message,
            )
            raise DataAccessError(_error_message) from _error
        except exc.InternalError as _error:
            print("postgresql error: {}".format(_error.orig.pgcode))
            self.session.rollback()
        except (exc.DataError, exc.IntegrityError, exc.StatementError) as _error:
            # This exception is raised when there is an error during
            # execution of a SQL statement.  These types of errors are
            # unlikely to be user errors as the programmer should ensure
            # everything is ready to insert.  Some situations that can raise
            # this exception are:
            #   1. Primary key violations.
            #   2. Non-date data supplied to date type fields.
            #   3. Foreign key violations.
            #   4. np.nan data supplied to any field type.
            #
            # With psycopg2, _error will have attributes pgcode and pgerror.
            # The first is a code associated with the error captured by
            # psycopg2 and the second is the original error message from the
            # database.  These should be used to generate the error message
            # to send to the client.  Error codes are defined in the
            # errorcodes.py file in the psycopg2 code base.
            self.session.rollback()
            _error_message = (
                "do_insert: Database error when attempting to add a record.  "
                "Database returned:\n\t{0:s}".format(
                    str(_error.orig.pgerror.split(":")[2].strip())
                )
            )
            pub.sendMessage(
                "fail_insert_record",
                error_message=_error_message,
            )
            raise DataAccessError(_error_message) from _error
        except FlushError as _error:
            self.session.rollback()
            _error_message = (
                "do_insert: Flush error when attempting to add records.  "
                "Database returned:\n\t{0:s}".format(str(_error))
            )
            pub.sendMessage(
                "fail_insert_record",
                error_message=_error_message,
            )
            raise DataAccessError(_error_message) from _error

    def do_insert_many(self, records: List[object]) -> None:
        """Add a group of new records to a database table.

        :param list records: the list of objects to add to the RAMSTK database.
        :return: None
        :rtype: None
        """
        for _record in records:
            self.do_insert(_record)

    def do_select_all(self, table, **kwargs) -> query.Query:
        """Select all records from the RAMSTK database for table.

        :param table: the database table object to select all from.
        :return: a list of table instances; one for each record.
        """
        _keys: List[str] = kwargs.get("key", None)
        _values: List[Any] = kwargs.get("value", None)
        _order: Any = kwargs.get("order", None)
        _all: bool = kwargs.get("_all", True)

        _filters = {}
        if _values[0] is not None:
            for _idx, _key in enumerate(_keys):
                _filters[_key] = _values[_idx]

        _results = self.session.query(table).filter_by(**_filters)
        if isinstance(_order, list):
            _results = _results.order_by(*_order)
        else:
            _results = _results.order_by(_order)

        if _all:
            _results = _results.all()
        else:
            _results = _results.first()

        return _results

    def do_update(self, record: object = None) -> None:
        """Update the RAMSTK database with any pending changes.

        :keyword record: the record to update in the database.
        :return: None
        :rtype: None
        """
        if record is not None:
            self.session.add(record)

        try:
            self.session.commit()
        except (
            exc.DataError,
            exc.IntegrityError,
            exc.InvalidRequestError,
            exc.ProgrammingError,
        ) as _error:
            self.session.rollback()
            _error_message = (
                "There was an database error when attempting to update a "
                "record.  Faulty SQL statement was:\n\t{0:s}.\nParameters "
                "were:\n\t{1:s}.".format(
                    str(_error.statement), str(_error.params)  # type: ignore
                )
            )  # type: ignore
            pub.sendMessage("fail_update_record", error_message=_error_message)
            raise DataAccessError(_error_message) from _error

    def get_database_list(self, database: Dict[str, str]) -> List:
        """Retrieve the list of program databases available to RAMSTK.

        This method is used to create a user-selectable list of databases when
        using the postgresql or MariaDB (MySQL) backend.  SQLite3 simply uses
        an open file dialog.

        :param database: the connection information for the dialect's
            administrative database.
        :return: the list of databases available to RAMSTK for the selected
            dialect.
        :rtype: list
        """
        _databases = []

        if database["dialect"] == "postgres":
            _query = self.sqlstatements["select"].format(
                "datname"
            ) + self.sqlstatements["from"].format("pg_database;")
            _database: str = (
                "postgresql+psycopg2://"
                + database["user"]
                + ":"
                + database["password"]
                + "@"
                + database["host"]
                + ":"
                + database["port"]
                + "/"
                + database["database"]
            )
            # pylint: disable=unused-variable
            __, _session = do_open_session(_database)

            # Make list of available databases, but only those associated with
            # RAMSTK.
            for db in _session.execute(_query):
                if db[0] not in ["postgres", "template0", "template1"]:
                    _databases.append(db[0])

        return _databases

    def get_last_id(self, table: str, id_column: str) -> Any:
        """Retrieve the last used value of the ID column.

        .. hint:: This method could be used to select the last used value from
            any column in a table.

        :param table: the name of the table to get the last ID from.
        :param id_column: the name of the field to use as the ID column.
        :return: _last_id; the last used value of the ID column.
        :rtype: int
        :raise: :class:`sqlalchemy.exc.OperationalError` if passed an unknown
            table or unknown column name.
        """
        # This ensures the column name is prefixed with fld_ in case the
        # table's attribute name was passed instead.
        if id_column[0:4] != "fld_":
            id_column = "fld_" + id_column

        _sql_statement = (
            self.sqlstatements["select"].format(id_column)
            + self.sqlstatements["from"].format(table)
            + self.sqlstatements["order"].format(id_column)
        )

        try:
            _last_id = self.session.execute(_sql_statement).first()[0]
        except (exc.ProgrammingError, TypeError):
            _last_id = 0

        return _last_id



# Standard Library Imports
from typing import Any, Dict

# RAMSTK Package Imports
from ramstk.views.gtk3.widgets import RAMSTKEntry as RAMSTKEntry
from ramstk.views.gtk3.widgets import RAMSTKFixedPanel as RAMSTKFixedPanel
from ramstk.views.gtk3.widgets import RAMSTKLabel as RAMSTKLabel

class MilHdbk217FResultPanel(RAMSTKFixedPanel):
    _record_field: str
    _select_msg: str
    _tag: str
    lblModel: RAMSTKLabel
    txtLambdaB: RAMSTKEntry
    txtPiQ: RAMSTKEntry
    txtPiE: RAMSTKEntry
    _hazard_rate_method_id: int
    _lambda_b: float
    category_id: int
    subcategory_id: int
    def __init__(self) -> None: ...
    def do_load_entries(self, attributes: Dict[str, Any]) -> None: ...
    def _do_set_hardware_attributes(self, attributes: Dict[str, Any]) -> None: ...
    def _do_set_reliability_attributes(self, attributes: Dict[str, Any]) -> None: ...
    def __do_set_model_label(self) -> None: ...



# pylint: disable=unused-import
"""The MIL-HDBK-217F Analyses Package."""

# RAMSTK Local Imports
from . import milhdbk217f
from .models import (
    capacitor,
    connection,
    crystal,
    efilter,
    fuse,
    inductor,
    integratedcircuit,
    lamp,
    meter,
    relay,
    resistor,
    semiconductor,
    switch,
)
